"""Definition file for observation data types"""

# each obsType should have TWO routines defined: _fielddata and _modelresult apended
# _fielddata(geo,dat,userEntry,obsDefault)
# _modelresult(geo,dat,lst,userEntry,obsDefault)

# userEntry is where input file (a list of observation specificaitions)
# each has obsType,obsInfo,fieldDataFile,customFilter,obsDefault
#
# obsInfo and fieldDataFile unique for each obs entry
# customFilter and obsDefault can be set or defaulted, mostly passed down
#     between onb entries unless modified.
#
# obsInfo is a list of items, specified as first line of each obs entry
#     it can be anything, eg. [string of name, number of time]
# fieldDataFile is a single string of ilename that the fielddata obs will be read from
#
# customFilter is a string that can be evaluated by the code to eg. limit range of
#     data etc
# obsDefault is an PestObservData object, the .OBSNME is the basename that should be
#     used for all created obs, other properties used as default.
#
# Each Observation is specified like this:
# [Obs]
# 1, 2, 3, 4, 5
# a_file_name.dat
#
# [Obs]
# 'abc'
# a_file_name.dat
# another_file.dat
#
# the first line (obsInfo) is always parsed by python, either as a list or a single value (of any type)
# the second and following line (fieldDataFile) are processed one by one, and will create different user entries.



# AY, May 2015
#
# I have done something major to allow the more flexible ways of specifying each observation.

from gopest.common import config as cfg

if 'waiwera' in cfg['simulator']['executable']:
    sim = 'waiwera'
else:
    sim = 'aut2'
FIELD = {
    'aut2': {
        'temp': 'Temperature',
        'pres': 'Pressure',
        'pco2': 'CO2 partial pressure',
        'rate': 'Generation rate',
        'enth': 'Enthalpy',
    },
    'waiwera':  {
        'temp': 'fluid_temperature',
        'pres': 'fluid_pressure',
        'pco2': 'fluid_CO2_partial_pressure',
        'rate': 'source_rate',
        'enth': 'source_enthalpy',
    },
}[sim]


PLOT_RAW_FIELD_DATA = True

shortNames = {
'Ex' : 'external',     # get obs value(s) from external source, thru json file
'En' : 'enthalpy',
'Ej' : 'enthalpy_json',
'Eb' : 'boiling',
'Bj' : 'boiling_json',
'Pr' : 'pressure',
'Pw' : 'pressure_by_well',
'Pb' : 'pressure_block_average',
'Pj' : 'pressure_block_average_json',
'Tw' : 'temperature',
'Th' : 'temperature_thickness',
'Tj' : 'temperature_thickness_json',
'Ti' : 'temp_interp_thickness_json',
'Tb' : 'blocktemperature',
'Hf' : 'heatflow',
'Hm' : 'heatflowminimum',
'Uf' : 'totalupflow',
'Ht' : 'totalheat',
'Se' : 'target_time',
}

# this for unique obs name
obsBaseNameCount = {}

def unique_obs_name(type_name, base):
    """ Form a unique observation name.  It will usually endup like
    XX_YYYYY_0012.  XX is shortNames[type_name], YYYYY is based on base and
    trimmed to 5 chars long.

    PEST max length of obs name is 20 chars.  If type_name is not one of the
    keys in shorNames, it will be used directly after trim to length of two.
    TODO: it's a bit ugly here, work on it!
    """
    base1 = type_name
    for s,t in shortNames.items():
        if t == type_name:
            base1 = s
    if base1 == type_name:
        base1 = type_name[:2]

    from gopest.common import private_cleanup_name
    base2 = private_cleanup_name(base)[:5]

    baseName = base1 +'_'+ base2
    if baseName not in obsBaseNameCount:
        obsBaseNameCount[baseName] = 0
    obsBaseNameCount[baseName] += 1
    if obsBaseNameCount[baseName] > 9999:
        raise Exception("Time to improve unique_obs_name()!")
    return baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])

def _matchInputGeners(dat, reg_exp_list, gener_types):
    """ return a list of (name matched) actual generators (objects)

    Matches is done using a list of re objects (reg_exp_list).  And the gener
    must have type in the specified tpye list.  Fixed/unfixed names will be
    dealt with properly.
    """
    gs = []
    from mulgrids import unfix_blockname
    for g in dat.generatorlist:
        for r in reg_exp_list:
            if r.match(g.name) or r.match(unfix_blockname(g.name)):
                if g.type in gener_types:
                    gs.append(g)
                # already know this GENER is included, check next gener
    return gs

def external_fielddata(geo, dat, userEntry):
    """ This observation type is designed for observation value(s) that is not
    extractable from the .listing file that goPESTobs reads.  Hence user is
    responsible for getting the values externally, and then stores it in a json
    file for goPESTobs to read back in.

    [Obs]
    filename.json  # first line of entry is the json file storing the value(s)
    'data_key'     # the key name used to acces the json file
    140.0          # obs value (a single float?)

    [Obs]
    abc.json       # first line of entry is the json file storing the value(s)
    'data_name'    # the key name used to acces the json file
    5.0            # a list of values, len() must match from json
    7.0
    9.0

    [Obs]
    abc.json       # first line of entry is the json file storing the value(s)
    'data_name'    # the key name used to acces the json file
    1., 5.         # lines of tuple of two values
    2., 7.
    3., 9.
    # data from json file will be interp() to match these

    The motivation for this is the predictive run for Bacman case.  Where
    results from a second .listing file is required.  Instead of making
    goPESTobs reads multiple listing file, I decided to make it an external
    extraction.  Usually this kind of obs needs very specialised calculation and
    may not be all that re-usable.

    WIP, the need for reading multiple listing files is still needed, as
    projects like Lihir needs a whole series of runs, and all results from all
    runs may be important.  I will do this when I am doing big surgery in goPEST
    code next time.  Further investigation for best design is required.
    """
    import numpy as np
    from copy import deepcopy

    obsInfo = userEntry.obsInfo
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault
    key = eval(obsInfo[1])        # usually use string as data key

    obses = []
    if len(obsInfo) == 3:
        expected = eval(obsInfo[2])
        if not isinstance(expected, float) and isinstance(expected, int):
            raise Exception
        ### single values
        expected = float(expected)
        obs = deepcopy(obsDefault)
        obs.OBSNME = unique_obs_name(obs.OBSNME, key)
        obs.OBSVAL = float(expected)
        obses.append(obs)
    elif len(obsInfo) > 3:
        expected = [eval(line) for line in obsInfo[2:]]
        if isinstance(expected[0], tuple):
            # list of tuples
            for x,y in expected:
                obs = deepcopy(obsDefault)
                obs.OBSNME = unique_obs_name(obs.OBSNME, key)
                obs.OBSVAL = y
                obses.append(obs)
        elif isinstance(expected[0], float) or isinstance(expected[0], int):
            # list of values
            for y in expected:
                obs = deepcopy(obsDefault)
                obs.OBSNME = unique_obs_name(obs.OBSNME, key)
                obs.OBSVAL = y
                obses.append(obs)
        else:
            raise Exception
    return obses

def external_modelresult(geo, dat, lst, userEntry):
    import numpy as np
    import json
    obsInfo = userEntry.obsInfo
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault

    with open(obsInfo[0], 'r') as f:
        data = json.load(f)
    key = eval(obsInfo[1])        # usually use string as data key

    values = []
    if len(obsInfo) == 3:
        ### single values
        if not isinstance(data[key], float) and not isinstance(data[key], int):
            raise Exception
        values.append(float(data[key]))
    elif len(obsInfo) > 3:
        expected = [eval(line) for line in obsInfo[2:]]
        if isinstance(expected[0], tuple):
            ### x-y values
            if len(data[key][0]) != len(data[key][1]):
                msg = '%s [%s] does not not have matching xs,ys: ' % (obsInfo[0], key)
                msg += '%i != %i' % (len(data[key][0]), len(data[key][1]))
                raise Exception(msg)
            exs = [float(e[0]) for e in expected]
            vs = np.interp(exs, data[key][0], data[key][1])
            values += list(vs)
        elif isinstance(expected[0], float) or isinstance(expected[0], int):
            ### list of values
            if not isinstance(data[key], list) or len(data[key]) != len(expected):
                msg = '%s [%s] does not match dimension of expected: ' % (obsInfo[0], key)
                msg += str(expected)
                raise Exception(msg)
            values += [float(x) for x in data[key]]
        else:
            raise Exception
    return values

def target_time_fielddata(geo, dat, userEntry):
    obsInfo = userEntry.obsInfo
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault

    expected_value = float(eval(obsInfo[0]))

    obses = []
    from copy import deepcopy
    obs = deepcopy(obsDefault)
    obs.OBSNME = unique_obs_name(obs.OBSNME, 'Se')
    obs.OBSVAL = float(expected_value)
    obses.append(obs)
    return obses

def target_time_modelresult(geo, dat, lst, userEntry):
    return [lst.fulltimes[-1]]

def totalheat_fielddata(geo,dat,userEntry):
    """ Extracting the total heat from specified GENERs (in dat/input file) as
    observation.

    [ObservationType]
    totalheat

    [Defaults]
    OBSNME = 'Heat'
    OBGNME = 'HtTotal'

    # !!! Note here uses all upper cases, don't touch other ObservationTypes
    [Obs]
    2000.00
    'abc99'
    'bbb99'
    'ccc99'

    # here it means an observation that adds up all heat in HEAT geners 'abc99',
    # 'bbb99', 'ccc99' (from data/input file) and the target of this sum is
    # 2000.0 J/s
    """
    obsInfo = userEntry.obsInfo
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault

    expected_value = float(eval(obsInfo[0]))

    # check if entry matches anything
    import re
    res = [re.compile(eval(line)) for line in userEntry.obsInfo[1:]]
    gs = _matchInputGeners(dat, res, ['HEAT'])
    if len(gs) == 0:
        name = "'%s'" % ("','".join([r.pattern for r in res]))
        raise Exception("Specified gener names does not match any geners: %s" % name)

    if len(obsInfo) > 2:
        ap = '_'
    else:
        ap = ''

    obses = []
    from copy import deepcopy
    obs = deepcopy(obsDefault)
    obs.OBSNME = unique_obs_name(obs.OBSNME, eval(obsInfo[1]) + ap)
    obs.OBSVAL = float(expected_value)
    obses.append(obs)
    return obses

def totalheat_modelresult(geo,dat,lst,userEntry):
    #go through all mass geners and extract their rate, sum this and return the total value..
    import re
    res = [re.compile(eval(line)) for line in userEntry.obsInfo[1:]]
    gs = _matchInputGeners(dat, res, ['HEAT'])
    total = float(sum([g.gx for g in gs]))
    return [total]

def totalupflow_fielddata(geo,dat,userEntry):
    obsInfo = userEntry.obsInfo
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault

    expected_value = float(eval(obsInfo[0]))

    if len(obsInfo) > 2:
        ap = '_'
    else:
        ap = ''

    obses = []
    from copy import deepcopy
    obs = deepcopy(obsDefault)
    obs.OBSNME = unique_obs_name(obs.OBSNME, eval(obsInfo[1]) + ap)
    obs.OBSVAL = float(expected_value)
    obses.append(obs)
    return obses

def totalupflow_modelresult(geo,dat,lst,userEntry):
    #go through all mass geners and extract their rate, sum this and return the total value..
    import re
    from mulgrids import unfix_blockname,fix_blockname
    res = []
    for line in userEntry.obsInfo[1:]:
        name = eval(line)
        res.append(re.compile(name))
    # these matchese should use the unfixed blockname rules
    total = 0.0
    for g in dat.generatorlist:
        for r in res:
            if r.match(g.name) or r.match(unfix_blockname(g.name)):
                if g.type in ('MASS','COM1'):
                    total += g.gx
                # already know this GENER is included, check next gener
                break
    return [total]


def heatflowminimum_fielddata(geo,dat,userEntry):
    expected_value = float(eval(userEntry.obsInfo[1]))
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault
    zoneName = eval(userEntry.obsInfo[0])

    from gopest.common import private_cleanup_name
    baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(zoneName)[:5]
    if baseName not in obsBaseNameCount:
        obsBaseNameCount[baseName] = 0
    obses = []
    obsBaseNameCount[baseName] += 1
    from copy import deepcopy
    obs = deepcopy(obsDefault)
    obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
    obs.OBSVAL = expected_value
    obses.append(obs)
    return obses

def heatflowminimum_modelresult(geo,dat,lst,userEntry):
    expected_value = float(eval(userEntry.obsInfo[1]))
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault
    zoneName = eval(userEntry.obsInfo[0])

    minimum = expected_value

    import config
    cfg = config.config('get_surface_heatflowminimum.cfg')

    from get_surface_heatflow import get_surface_heatflow_proc_cfg
    (geners,colsinzones,ListingTableNames,syear,coldenthalpy,
        show_fig,save_fig,outflow_only,calc_notinany
        ) = get_surface_heatflow_proc_cfg(cfg,geo,lst)

    if zoneName not in colsinzones: raise Exception("'%s' not in colsinzones" % zoneName + str(sorted(colsinzones.keys())))
    # so to skip other zones, still ineffecient, but that's it for now
    for z in colsinzones.keys():
        if z != zoneName: del colsinzones[z]

    from get_surface_heatflow import get_surface_heatflow
    (t_in_sec, zone_total, zone_area) = get_surface_heatflow(geo,lst,
        geners,colsinzones,ListingTableNames,syear,coldenthalpy,
        False,False,outflow_only,calc_notinany)

    # Only NS total heatflow (result index [-1]).  If model result is greater or
    # equal to the expected_value, than return expected_value, so it fits
    # perfectly as long as model result is > expected_value.
    if list(zone_total[zoneName])[-1] >= minimum:
        return [minimum]
    else:
        return [list(zone_total[zoneName])[-1]]



def heatflow_fielddata(geo,dat,userEntry):
    expected_value = float(eval(userEntry.obsInfo[1]))
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault
    zoneName = eval(userEntry.obsInfo[0])

    from gopest.common import private_cleanup_name
    baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(zoneName)[:5]
    if baseName not in obsBaseNameCount:
        obsBaseNameCount[baseName] = 0
    obses = []
    obsBaseNameCount[baseName] += 1
    from copy import deepcopy
    obs = deepcopy(obsDefault)
    obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
    obs.OBSVAL = expected_value
    obses.append(obs)
    return obses

def heatflow_modelresult(geo,dat,lst,userEntry):
    expected_value = userEntry.obsInfo[1]
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault
    zoneName = eval(userEntry.obsInfo[0])

    import config
    cfg = config.config('get_surface_heatflow.cfg')

    from get_surface_heatflow import get_surface_heatflow_proc_cfg
    (geners,colsinzones,ListingTableNames,syear,coldenthalpy,
        show_fig,save_fig,outflow_only,calc_notinany
        ) = get_surface_heatflow_proc_cfg(cfg,geo,lst)

    if zoneName not in colsinzones: raise Exception
    # so to skip other zones, still ineffecient, but that's it for now
    for z in colsinzones.keys():
        if z != zoneName: del colsinzones[z]

    from get_surface_heatflow import get_surface_heatflow
    (t_in_sec, zone_total, zone_area) = get_surface_heatflow(geo,lst,
        geners,colsinzones,ListingTableNames,syear,coldenthalpy,
        False,False,outflow_only,calc_notinany)

    # just get NS total heatflow for now
    return [list(zone_total[zoneName])[-1]]


def private_well_track_blocks(geo,wname):
    """ generate well track blocks if not already cached """
    if 'wellblocks' not in geo.__dict__:
        geo.wellblocks = {}
    if wname in geo.wellblocks:
        return geo.wellblocks[wname]
    else:
        # spped up, hopefully
        if 'qtree' not in geo.__dict__:
            geo.qtree = geo.column_quadtree()

        # work out all blocks
        w = geo.well[wname]
        blocks, blocks_cen = [], []
        for lc in [lay.centre for lay in geo.layerlist if w.bottom[2] <= lay.centre <= w.head[2]]:
            b = geo.block_name_containing_point(w.elevation_pos(lc),geo.qtree)
            if b is not None:
                blocks.append(b)
                blocks_cen.append(lc)

        # check if well head block is missed
        wh_col = geo.column_containing_point(w.head[:2],
            guess=geo.column[geo.column_name(blocks[0])],qtree=geo.qtree)
        for lay in geo.layerlist:
            if lay.bottom < wh_col.surface:
                wh_block_elev = (wh_col.surface + lay.bottom ) / 2.0
                wh_block = geo.block_name(lay.name,wh_col.name)
                break
        if wh_block not in blocks:
            blocks.insert(0, wh_block)
            blocks_cen.insert(0, wh_block_elev)
        geo.wellblocks[wname] = (blocks, blocks_cen)
        return geo.wellblocks[wname]

def _loadBlockTempFile(fname, customFilter):
    """ get all block names and temp out of field data file """
    from mulgrids import fix_blockname
    allblks, alltemp = [], []
    f = open(fname,'r')
    for line in f.readlines():
        if line.strip() == '': break
        block,temp = [eval(x) for x in line.split(',')[0:2]]
        if eval(customFilter):
            allblks.append(fix_blockname(block))
            alltemp.append(float(temp))
    f.close()
    return allblks, alltemp

def blocktemperature_fielddata(geo,dat,userEntry):
    """ a user field data file is a list of blocks with oberved temperature """
    fieldDataFile = userEntry.obsInfo[1]
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault

    allblks, alltemp = _loadBlockTempFile(fieldDataFile, customFilter)

    from mulgrids import fix_blockname
    from copy import deepcopy
    obses = []
    for (b,t) in zip(allblks,alltemp):
        obs = deepcopy(obsDefault)
        obs.OBSNME = unique_obs_name(obsDefault.OBSNME, fix_blockname(b))
        obs.OBSVAL = t
        obses.append(obs)
    return obses

def blocktemperature_modelresult(geo,dat,lst,userEntry):
    """ a user field data file is a list of blocks with oberved temperature """
    fieldDataFile = userEntry.obsInfo[1]
    customFilter = userEntry.customFilter

    allblks, alltemp = _loadBlockTempFile(fieldDataFile, customFilter)

    vals = eval(userEntry.obsInfo[0])
    time = 0.0
    if isinstance(vals,tuple) and len(vals) == 2:
        time = float(vals[1])

    import numpy as np
    lst.index = np.abs(lst.fulltimes-time).argmin()

    field_name = [c for c in lst.element.column_name if c.startswith(FIELD['temp'])][0]

    return [lst.element[b][field_name] for b in allblks]

def temperature_fielddata(geo,dat,userEntry):
    # ugly, need re-writting and remove repeative actions
    fieldDataFile = userEntry.obsInfo[1]
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault
    from mulgrids import fix_blockname
    vals = eval(userEntry.obsInfo[0])
    if isinstance(vals,str):
        # only wellname is specified
        wname = fix_blockname(vals)
    elif isinstance(vals,tuple):
        wname = fix_blockname(vals[0])

    # get temp vs elev from datafile first
    allelev, alltemp = [], []
    f = open(fieldDataFile,'r')
    for line in f.readlines():
        if line.strip() == '': break
        elev,temp = [float(x) for x in line.split()[0:2]]
        if eval(customFilter):
            allelev.append(elev)
            alltemp.append(temp)
    f.close()

    (bs, bs_c) = private_well_track_blocks(geo,wname)
    blocks, blocks_cen = [], []
    for (b,c) in zip(bs, bs_c):
        if allelev[-1] <= c <= allelev[0]:
            blocks.append(b)
            blocks_cen.append(c)

    # blocks and blocks_cen ready for use
    from numpy import interp
    blocks_temp = interp(blocks_cen,allelev[::-1],alltemp[::-1])

    from gopest.common import private_cleanup_name
    baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(wname)[:5]
    if baseName not in obsBaseNameCount:
        obsBaseNameCount[baseName] = 0
    obses = []
    for (b,t) in zip(blocks,blocks_temp):
        obsBaseNameCount[baseName] += 1
        from copy import deepcopy
        obs = deepcopy(obsDefault)
        obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
        obs.OBSVAL = t
        obses.append(obs)
    return obses

def temperature_modelresult(geo,dat,lst,userEntry):
    # ugly, need re-writting and remove repeative actions
    # maybe communication between pre- and post - process is still needed
    fieldDataFile = userEntry.obsInfo[1]
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault
    from mulgrids import fix_blockname
    vals = eval(userEntry.obsInfo[0])
    time = 0.0
    if isinstance(vals,str):
        # only wellname is specified
        wname = fix_blockname(vals)
    elif isinstance(vals,tuple):
        wname = fix_blockname(vals[0])
        if len(vals) == 2:
            time = float(vals[1])

    # get temp vs elev from datafile first
    allelev, alltemp = [], []
    f = open(fieldDataFile,'r')
    for line in f.readlines():
        if line.strip() == '': break
        elev,temp = [float(x) for x in line.split()[0:2]]
        if eval(customFilter):
            allelev.append(elev)
            alltemp.append(temp)
    f.close()

    (bs, bs_c) = private_well_track_blocks(geo,wname)
    blocks = []
    for (b,c) in zip(bs, bs_c):
        if allelev[-1] <= c <= allelev[0]:
            blocks.append(b)

    import numpy as np
    lst.index = np.abs(lst.fulltimes-time).argmin()

    field_name = [c for c in lst.element.column_name if c.startswith(FIELD['temp'])][0]
    return [lst.element[b][field_name] for b in blocks]

def temperature_thickness_fielddata(geo,dat,userEntry):
    """ This is very simialr to the norml temperature obs type, only that the
    weight of each observation is assigned as the thickness of that layer.  The
    observation actually seen by PEST is saved as .obs files. """
    # ugly, need re-writting and remove repeative actions
    fieldDataFile = userEntry.obsInfo[1]
    customFilter = userEntry.customFilter
    obsDefault = userEntry.obsDefault
    from mulgrids import fix_blockname
    vals = eval(userEntry.obsInfo[0])
    if isinstance(vals,str):
        # only wellname is specified
        wname = fix_blockname(vals)
    elif isinstance(vals,tuple):
        wname = fix_blockname(vals[0])

    # get temp vs elev from datafile first
    allelev, alltemp = [], []
    f = open(fieldDataFile,'r')
    for line in f.readlines():
        if line.strip() == '': break
        elev,temp = [float(x) for x in line.split()[0:2]]
        if eval(customFilter):
            allelev.append(elev)
            alltemp.append(temp)
    f.close()

    (bs, bs_c) = private_well_track_blocks(geo,wname)
    blocks, blocks_cen, blocks_thickness = [], [], []
    for (b,c) in zip(bs, bs_c):
        if allelev[-1] <= c <= allelev[0]:
            blocks.append(b)
            blocks_cen.append(c)
            blocks_thickness.append(geo.layer[geo.layer_name(b)].thickness)

    # blocks and blocks_cen ready for use
    from numpy import interp
    blocks_temp = interp(blocks_cen,allelev[::-1],alltemp[::-1])

    fo = open(fieldDataFile+'.obs', 'w')
    for (z,t) in zip(blocks_cen,blocks_temp):
        fo.write('%.2f %.2f\n' % (z,t))
    fo.close()

    from gopest.common import private_cleanup_name
    baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(wname)[:5]
    if baseName not in obsBaseNameCount:
        obsBaseNameCount[baseName] = 0
    obses = []


    if obsDefault.OBGNME not in userEntry.coverage:
        userEntry.coverage[obsDefault.OBGNME] = []
    for (b,t,h) in zip(blocks,blocks_temp, blocks_thickness):
        obsBaseNameCount[baseName] += 1
        from copy import deepcopy
        obs = deepcopy(obsDefault)
        obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
        obs.OBSVAL = t
        obs.WEIGHT = h * obsDefault.WEIGHT
        userEntry.coverage[obsDefault.OBGNME].append(b)
        obses.append(obs)
    return obses

def temperature_thickness_modelresult(geo,dat,lst,userEntry):
    return temperature_modelresult(geo,dat,lst,userEntry)

def _well_interp_layers(wname, geo):
    """
    WARNING: this can be very problematic at the surface layers near wellhead.
    There would be cases where blocks involved is partial layer, then the
    interpolation would wrongly use layer centre as interpolation in 2D.
    """
    well = geo.well[wname]
    # cache qtree/kdtree for faster repeated searching
    if not hasattr(geo, 'kdtree'): geo.kdtree = geo.get_node_kdtree()
    if not hasattr(geo, 'qtree'): geo.qtree = geo.column_quadtree()

    interp_block_i, well_poses, elevs = [], [], []
    for lay in geo.layerlist[1:]:
        pos = well.elevation_pos(lay.centre, extend=False)
        if pos is None:
            # skip, well pos above or below well track
            continue

        nearest_node = geo.node_nearest_to(pos[:2], kdtree=geo.kdtree)
        well_col = geo.column_containing_point(pos[:2], qtree=geo.qtree)
        well_bname = geo.block_name_containing_point(pos, qtree=geo.qtree)

        if well_bname is None:
            # well block not in model, skip layer
            continue

        blocks_i = [] # block index for slicing listing table
        points = []
        for col in nearest_node.column | well_col.neighbour:
            bname = geo.block_name(lay.name, col.name)
            try:
                blocks_i.append(geo.block_name_index[bname])
            except KeyError:
                # incomplete layer, just use well block
                interp_block_i.append([geo.block_name_index[well_bname]])
                well_poses.append(pos)
                elevs.append(lay.centre)
                # next layer
                continue

        interp_block_i.append(blocks_i)
        well_poses.append(pos)
        elevs.append(lay.centre)
    return interp_block_i, well_poses, elevs

def _well_interp_temp(block_i_by_layer, well_poses, geo, lst):
    """ block_i_by_layer is a list of lists, each a group of block indices to
    interpolate temperature data.  well_poses is a list of pos, one for each
    layers same as block_i_by_layer.
    """
    import scipy.interpolate as interpolate
    # will cause error if no field name found
    for field in lst.element.column_name:
        if field.startswith('Temperatu'):
            field_name = field
            break
    temps = []
    for blocks_i, pos in zip(block_i_by_layer, well_poses):
        v = lst.element[field_name][blocks_i]
        cols = [geo.column_name(geo.block_name_list[bi]) for bi in blocks_i]
        points = [geo.column[c].centre for c in cols]
        if len(points) > 1:
            vi = interpolate.griddata(points, v, [pos[:2]], method='linear')
            temps.append(vi[0])
        else:
            temps.append(v[0])
    return temps


def temp_interp_thickness_json_fielddata(geo,dat,userEntry):
    jfilename = userEntry.obsInfo[0]
    customFilter = userEntry.customFilter
    offsetTime = userEntry.offsetTime
    obsDefault = userEntry.obsDefault
    tFactor = 365.25*24.*60.*60.

    import json
    # 1st line is json file name (of all wells)
    with open(jfilename, 'r') as f:
        t_bywell = json.load(f)

    obses = []

    for oline in userEntry.obsInfo[1:]:
        wname = eval(oline)
        if 'geo_well_name' in t_bywell[wname]:
            geo_wname = t_bywell[wname]['geo_well_name']
        else:
            geo_wname = wname
        time = t_bywell[wname]['time']
        allelev, alltemp = [], []
        for elev,temp in zip(t_bywell[wname]['elevations'],
                             t_bywell[wname]['temperatures']):
            if eval(customFilter):
                allelev.append(elev)
                alltemp.append(temp)

        # bis, wps, es = blk indices, well positions, elevations
        bis, wps, es = _well_interp_layers(geo_wname, geo)
        blocks, blocks_cen, blocks_thickness = [], [], []
        bis2, wps2, es2, bhs2 = [], [], [], []
        for bidx, wpos, elev in zip(bis, wps, es):
            if elev > allelev[0] or elev < allelev[-1]:
                # skip if out of field data range
                continue
            if eval(customFilter):
                bis2.append(bidx)
                wps2.append(wpos)
                es2.append(elev)
                # all same layer, just use first one
                lay = geo.layer_name(geo.block_name_list[bidx[0]])
                bhs2.append(geo.layer[lay].thickness)

        # blocks and blocks_cen ready for use
        from numpy import interp
        blocks_temp = interp(es2, allelev[::-1],alltemp[::-1])

        from gopest.common import private_cleanup_name
        baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(geo_wname)[:5]
        if baseName not in obsBaseNameCount:
            obsBaseNameCount[baseName] = 0

        if obsDefault.OBGNME not in userEntry.coverage:
            userEntry.coverage[obsDefault.OBGNME] = []
            userEntry.coverage[obsDefault.OBGNME+'_interp_source'] = []



        for bidx,pos,t,h in zip(bis2, wps2, blocks_temp, bhs2):
            obsBaseNameCount[baseName] += 1
            from copy import deepcopy
            obs = deepcopy(obsDefault)
            obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
            obs.OBSVAL = t
            obs.WEIGHT = h * obsDefault.WEIGHT
            # additional for model result extraction
            obs._bindx_ = bidx
            obs._wpos_ = pos
            obs._dtime_ = time * tFactor - offsetTime # data's time tag
            blks = [geo.block_name_list[i] for i in bidx]
            userEntry.coverage[obsDefault.OBGNME] = userEntry.coverage[obsDefault.OBGNME] + blks
            userEntry.coverage[obsDefault.OBGNME+'_interp_source'].append((
                obs.OBSNME, bidx))
            obses.append(obs)

        # generate batch plot entries
        plot = temperature_plot(
            geo_wname, time, list(es2), list(blocks_temp),
            baseName, ('Well %s at time %f' % (wname,time)) )
        if PLOT_RAW_FIELD_DATA:
            y, x = t_bywell[wname]['elevations'], t_bywell[wname]['temperatures']
            plot = plot_append_raw_data(plot, 'raw_'+wname, x, y, xunit="degC", yunit="meter")
        userEntry.batch_plot_entry.append(plot)

    return obses

def temp_interp_thickness_json_modelresult(geo,dat,lst,userEntry):
    import numpy as np
    import scipy.interpolate as interpolate
    # will cause error if no field name found
    for field in lst.element.column_name:
        if field.startswith(FIELD['temp']):
            field_name = field
            break
    print('+++ use field: %s' % field_name)

    obses = temp_interp_thickness_json_fielddata(geo, dat, userEntry)
    vals = []

    t_prev = obses[0]._dtime_
    lst.index = np.abs(lst.fulltimes-t_prev).argmin()
    for obs in obses:
        bis = obs._bindx_
        wps = obs._wpos_
        t = obs._dtime_

        # TODO, this is slow, can be much faster
        if t != t_prev:
            lst.index = np.abs(lst.fulltimes-t).argmin()
            t_prev = t

        v = lst.element[field_name][bis]
        cols = [geo.column_name(geo.block_name_list[bi]) for bi in bis]
        points = [geo.column[c].centre for c in cols]
        if len(points) > 1:
            vi = interpolate.griddata(points, v, [wps[:2]], method='linear')
            vals.append(vi[0])
        else:
            vals.append(v[0])
    return vals


def temperature_thickness_json_fielddata(geo,dat,userEntry):
    """ This is very simialr to the norml temperature obs type, only that the
    weight of each observation is assigned as the thickness of that layer.  The
    observation actually seen by PEST is saved as .obs files. """
    # ugly, need re-writting and remove repeative actions
    jfilename = userEntry.obsInfo[0]
    customFilter = userEntry.customFilter
    offsetTime = userEntry.offsetTime
    obsDefault = userEntry.obsDefault
    tFactor = 365.25*24.*60.*60.

    import json
    # 1st line is json file name (of all wells)
    with open(jfilename, 'r') as f:
        t_bywell = json.load(f)

    obses = []

    for oline in userEntry.obsInfo[1:]:
        wname = eval(oline)
        if 'geo_well_name' in t_bywell[wname]:
            geo_wname = t_bywell[wname]['geo_well_name']
        else:
            geo_wname = wname
        time = t_bywell[wname]['time']
        allelev, alltemp = [], []
        for elev,temp in zip(t_bywell[wname]['elevations'], t_bywell[wname]['temperatures']):
            if eval(customFilter):
                allelev.append(elev)
                alltemp.append(temp)

        if len(allelev) <= 1:
            print('temperature_thickness_json: User entry has no data, skipping: %s' % wname)
            continue

        (bs, bs_c) = private_well_track_blocks(geo,geo_wname)
        blocks, blocks_cen, blocks_thickness = [], [], []
        for (b,c) in zip(bs, bs_c):
            if allelev[-1] <= c <= allelev[0]:
                blocks.append(b)
                blocks_cen.append(c)
                blocks_thickness.append(geo.layer[geo.layer_name(b)].thickness)

        # blocks and blocks_cen ready for use
        from numpy import interp
        blocks_temp = interp(blocks_cen,allelev[::-1],alltemp[::-1])

        from gopest.common import private_cleanup_name
        baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(geo_wname)[:5]
        if baseName not in obsBaseNameCount:
            obsBaseNameCount[baseName] = 0

        if obsDefault.OBGNME not in userEntry.coverage:
            userEntry.coverage[obsDefault.OBGNME] = []
        for (b,t,h) in zip(blocks,blocks_temp, blocks_thickness):
            obsBaseNameCount[baseName] += 1
            from copy import deepcopy
            obs = deepcopy(obsDefault)
            obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
            obs.OBSVAL = t
            obs.WEIGHT = h * obsDefault.WEIGHT
            # additional for model result extraction
            obs._block_ = b
            obs._dtime_ = time * tFactor - offsetTime # data's time tag
            userEntry.coverage[obsDefault.OBGNME].append(b)
            obses.append(obs)

        # generate batch plot entries
        plot = temperature_plot(
            geo_wname, time, blocks_cen, list(blocks_temp),
            baseName, ('Well %s at time %f' % (wname,time)) )
        if PLOT_RAW_FIELD_DATA:
            y, x = t_bywell[wname]['elevations'], t_bywell[wname]['temperatures']
            plot = plot_append_raw_data(plot, 'raw_'+wname, x, y, xunit="degC", yunit="meter")
        userEntry.batch_plot_entry.append(plot)
        # generate batch plot entries
        # userEntry.batch_plot_entry.append(temperature_plot(
        #     wname, time, allelev, alltemp,
        #     baseName, ('Well %s at time %f' % (wname,time)) ))

    return obses

def temperature_thickness_json_modelresult(geo,dat,lst,userEntry):
    import numpy as np
    obses = temperature_thickness_json_fielddata(geo, dat, userEntry)
    vals = []
    t_prev = obses[0]._dtime_
    lst.index = np.abs(lst.fulltimes-t_prev).argmin()
    for obs in obses:
        b = obs._block_
        t = obs._dtime_

        # TODO, this is slow, can be much faster
        if t != t_prev:
            lst.index = np.abs(lst.fulltimes-t).argmin()
            t_prev = t

        vals.append(lst.element[b][FIELD['temp']])
    return vals

def temperature_plot(wname, time, elevs, temps, obsname, title):
    """ for (timgui) batch plotting """
    return {
        "series": [
            {
                "variable": FIELD["temp"],
                "well": wname,
                "timeunit": "year",
                "type": "DownholeWellSeries",
                "time": time
            },
            {
                "frozen_x": temps,
                "frozen_y": elevs,
                "xunit": "degC",
                "yunit": "meter",
                "name": obsname,
                "original_series": "",
                "type": "FrozenDataSeries"
            }
        ],
        "ylabel": "Elevation",
        "xlabel": "Temperature",
        "title": title
    }


def pressure_fielddata(geo,dat,userEntry):
    """ called by goPESTobs.py """
    entries = private_history_data(userEntry, 100000.0, 365.25*24.*60.*60.)
    obses, times = zip(*entries)
    return obses
def pressure_modelresult(geo,dat,lst,userEntry):
    from mulgrids import fix_blockname
    # name,timelist
    name = fix_blockname(eval(userEntry.obsInfo[0]))
    entries = private_history_data(userEntry, 100000.0, 365.25*24.*60.*60.)
    obses, timelist = zip(*entries)
    tbl = lst.history([('e',name,'Pressure')])
    if tbl is None:
        raise Exception("Observation (type pressure) '%s' does not match any block." % name)
    alltimes = tbl[0] # assuming all times are the same
    allpress = tbl[1]
    from numpy import interp
    return list(interp(timelist,alltimes,allpress))

def pressure_by_well_fielddata(geo,dat,userEntry):
    """ called by goPESTobs.py """
    entries = private_history_data(userEntry, 100000.0, 365.25*24.*60.*60.)
    if len(entries) == 0:
        raise Exception("User entry yields no observation: " + str(userEntry))
    obses, times = zip(*entries)
    return obses
def pressure_by_well_modelresult(geo,dat,lst,userEntry):
    """ expects a well name and elevation in first line, eg: 'WK  1', -100.0 """
    from mulgrids import fix_blockname
    # name,timelist
    # name = fix_blockname(eval(userEntry.obsInfo[0]))
    wname, elev = eval(userEntry.obsInfo[0])
    elev = float(elev)
    if wname not in geo.well:
        raise Exception("Obs type 'pressure_by_well' well %s does not exist in geometry file." % wname)
    pos = geo.well[wname].elevation_pos(elev, extend=True)
    # spped up, hopefully
    if 'qtree' not in geo.__dict__:
        geo.qtree = geo.column_quadtree()
    name = geo.block_name_containing_point(pos, geo.qtree)
    if name is None:
        raise Exception("Obs type 'pressure_by_well' well %s at %f is outside of the model." % (wname, elev))

    # print wname, elev, name
    entries = private_history_data(userEntry, 100000.0, 365.25*24.*60.*60.)
    obses, timelist = zip(*entries)
    tbl = lst.history([('e',name,'Pressure')])
    if tbl is None:
        raise Exception("Obs failed to extract Pressure for block %s." % name)
    alltimes = tbl[0] # assuming all times are the same
    allpress = tbl[1]
    from numpy import interp
    return list(interp(timelist,alltimes,allpress))

def target_times(desired_times, limit, data):
    """ work out target times (among desired_times) that has data within +-
    limit """
    def count_valid(time, limit, data):
        """ Assuming data is a list of tuple (t,y), count how many t is
        within the range of time +- limit. """
        i = 0
        for t,y in data:
            if (time-limit) <= t <= (time+limit):
                i += 1
        return i
    targets = []
    for time in desired_times:
        if count_valid(time, limit, data):
            targets.append(time)
    return targets

def calc_bz(geo, w, elev):
    pos = geo.well[w].elevation_pos(elev, extend=True)
    if 'qtree' not in geo.__dict__:
        geo.qtree = geo.column_quadtree()
    b = geo.block_name_containing_point(pos, geo.qtree)
    if b is None:
        raise Exception("well %s at %f is outside of the model." % (w, pcp))
    return b, geo.block_centre(geo.layer_name(b), geo.column_name(b))[2]

def pressure_plot(bname, dataname, ts, ps, title):
    """ for (timgui) batch plotting """
    return {
        "series": [
            {
                "type": "HistoryBlockSeries",
                "block": bname,
                "variable": FIELD["pres"]
            },
            {
                "type": "FrozenDataSeries",
                "name": dataname,
                "original_series": "",
                "frozen_x": [t for t in ts],
                "frozen_y": [p*1.0e5 for p in ps],
                "xunit": "t2year",
                "yunit": "pascal",
            }
        ],
        "ylabel": "Pressure",
        "xlabel": "Time",
        "title": title
    }

def plot_append_raw_data(plot, name, x, y, xunit="", yunit=""):
    plot["series"].append({
                              "type": "FrozenDataSeries",
                              "name": name,
                              "original_series": "",
                              "frozen_x": x,
                              "frozen_y": y,
                              "xunit": xunit,
                              "yunit": yunit,
                          })
    return plot

def pressure_block_average_fielddata(geo, dat, userEntry):
    customFilter = userEntry.customFilter
    offsetTime = userEntry.offsetTime
    obsDefault = userEntry.obsDefault
    vFactor, tFactor = 100000.0, 365.25*24.*60.*60.

    # check all required settings exists:
    must_have = ['_DESIRED_DATA_TIMES', '_INTERP_LIMIT', '_P_GRADIENT']
    if any([not hasattr(obsDefault,a) for a in must_have]):
        raise Exception('Obs type pressure_block_average must have these default settings: ' + ', '.join(must_have))

    p_byblock = {}

    for oline in userEntry.obsInfo:
        wname, elev, fwell = eval(oline)
        elev = float(elev)

        f = open(fwell,'r')
        times, vals = [], []
        for line in f.readlines():
            if line.strip() == '': break
            time,val = [float(x) for x in line.split()[0:2]]
            if eval(customFilter):
                times.append(time), vals.append(val)
        f.close()

        if len(times) == 0:
            f.close()
            raise Exception("Pressure file: %s yields no observation" % fwell)

        desired_times = obsDefault._DESIRED_DATA_TIMES
        interp_limit = obsDefault._INTERP_LIMIT
        p_gradient = obsDefault._P_GRADIENT

        # correct to block centre
        b, bz = calc_bz(geo, wname, elev)
        vals = [p - (bz-elev) * p_gradient for p in vals]

        # get times and vals for each line (well)
        final_times = [t for t in desired_times if times[0] <= t <= times[-1]]
        final_times = target_times(desired_times, interp_limit, zip(times,vals))
        if len(final_times) == 0:
            raise Exception("User entry yields no observation: " + oline)
        from numpy import interp
        final_vals = list(interp(final_times,times,vals))

        if b not in p_byblock:
            p_byblock[b] = []
        # each block have a list (ts, ps), one for each well in block
        p_byblock[b].append((final_times, final_vals, wname))

    # one series of obs for each block, averaging all these in the same block
    obses = []
    for b in sorted(p_byblock.keys()):
        # fo = open('p_byblock_%s.obs' % b.replace(' ','_'), 'w')
        bp_times = {}
        ws = []
        for dtimes, dvals, wname in p_byblock[b]:
            ws.append(wname)
            for ft,fv in zip(dtimes, dvals):
                if ft not in bp_times:
                    bp_times[ft] = []
                bp_times[ft].append(fv)

        avgt, avgp = [], []
        for t in sorted(bp_times.keys()):
            avgt.append(t)
            avgp.append(sum(bp_times[t]) / len(bp_times[t]))

        from gopest.common import private_cleanup_name
        from copy import deepcopy
        baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(b)
        if baseName not in obsBaseNameCount:
            obsBaseNameCount[baseName] = 0
        for time, val in zip(avgt, avgp):
            obsBaseNameCount[baseName] += 1

            obs = deepcopy(obsDefault)
            obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
            obs.OBSVAL = val * vFactor
            # additional for model result extraction
            obs._block_ = b
            obs._mtime_ = time * tFactor - offsetTime
            obses.append(obs)
            # output data file in original unit (instead of PEST/TOUGH2)
            # fo.write('%e %e\n' % (time, val))

        # generate batch plot entries
        userEntry.batch_plot_entry.append(pressure_plot(
            b, baseName, [t-offsetTime/(60.0*60.0*24.0*365.25) for t in avgt], avgp,
            b+': '+','.join(ws)))

        # fo.close()
    return obses

def pressure_block_average_modelresult(geo,dat,lst,userEntry):
    obses = pressure_block_average_fielddata(geo, dat, userEntry)
    bs, tss = [], []
    for obs in obses:
        b = obs._block_
        t = obs._mtime_
        # print b, t
        if b not in bs:
            bs.append(b)
            tss.append([])
        tss[-1].append(t)

    tbls = lst.history([('e',b,'Pressure') for b in bs])
    if tbls is None:
        raise Exception("Extraction Pressure history of listing results went wrong, blocks: %s" % ','.join(bs))
    # PyTOUGH checks result length, then just return the first set (hence a
    # tuple) if only one in the list, instead of a list
    if not isinstance(tbls, list):
        tbls = [tbls]

    from numpy import interp
    all_obs_vals = []
    for i, (b,ts) in enumerate(zip(bs,tss)):
        alltimes = tbls[i][0] # assuming all times are the same
        allpress = tbls[i][1]
        all_obs_vals += list(interp(ts,alltimes,allpress))

    return all_obs_vals

def pressure_block_average_json_fielddata(geo, dat, userEntry):
    customFilter = userEntry.customFilter
    offsetTime = userEntry.offsetTime
    obsDefault = userEntry.obsDefault
    vFactor, tFactor = 100000.0, 365.25*24.*60.*60.

    # check all required settings exists:
    must_have = ['_DESIRED_DATA_TIMES', '_INTERP_LIMIT', '_P_GRADIENT']
    if any([not hasattr(obsDefault,a) for a in must_have]):
        raise Exception('Obs type pressure_block_average_json must have these default settings: ' + ', '.join(must_have))

    import json
    # 1st line is json file name (of all wells)
    with open(userEntry.obsInfo[0], 'r') as f:
        p_bywell = json.load(f)

    p_byblock = {}

    for oline in userEntry.obsInfo[1:]:
        wname = eval(oline)
        elev = p_bywell[wname]['elevation']

        ts, vs = p_bywell[wname]['times'], p_bywell[wname]['pressures']
        times, vals = [], []
        for time, val in zip(ts, vs):
            if eval(customFilter):
                times.append(time), vals.append(val)

        if len(times) == 0:
            f.close()
            raise Exception("Pressure from %s: %s yields no observation" % (userEntry.obsInfo[0], wname))

        desired_times = obsDefault._DESIRED_DATA_TIMES
        interp_limit = obsDefault._INTERP_LIMIT
        p_gradient = obsDefault._P_GRADIENT

        ##look for the column that contains the bottom of the well
        #for col in geo.columnlist:
        #    #position of well bottom
        #    pos = geo.well[wname].bottom
        #    if col.contains_point(pos):
        #        depth = col.surface - pos[3]
        #
        #geo.column_containing_point(geo.well[wname].bottom)

        # b, bz = calc_bz(geo, wname, pos[3])
        # vals = [p - (bz-pos[3]) * p_gradient for p in vals]

        # correct to block centre
        b, bz = calc_bz(geo, wname, elev)
        vals = [p - (bz-elev) * p_gradient for p in vals]

        # get times and vals for each line (well)
        final_times = [t for t in desired_times if times[0] <= t <= times[-1]]
        final_times = target_times(desired_times, interp_limit, list(zip(times,vals)))
        if len(final_times) == 0:
            msg1 = 'final_times = ' + str(final_times)
            msg2 = 'interp_limit = ' + str(interp_limit)
            msg3 = 'times = ' + str(times)
            msg = "User entry yields no observation: " + oline
            raise Exception('\n'.join([msg1, msg2, msg3, msg]))
        from numpy import interp
        final_vals = list(interp(final_times,times,vals))

        if b not in p_byblock:
            p_byblock[b] = []
        # each block have a list (ts, ps), one for each well in block
        p_byblock[b].append((final_times, final_vals, wname))

    # one series of obs for each block, averaging all these in the same block
    obses = []
    for b in sorted(p_byblock.keys()):
        # fo = open('p_byblock_%s.obs' % b.replace(' ','_'), 'w')
        bp_times = {}
        ws = []
        for dtimes, dvals, wname in p_byblock[b]:
            ws.append(wname)
            for ft,fv in zip(dtimes, dvals):
                if ft not in bp_times:
                    bp_times[ft] = []
                bp_times[ft].append(fv)

        avgt, avgp = [], []
        for t in sorted(bp_times.keys()):
            avgt.append(t)
            avgp.append(sum(bp_times[t]) / len(bp_times[t]))

        from gopest.common import private_cleanup_name
        from copy import deepcopy
        baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(b)
        if baseName not in obsBaseNameCount:
            obsBaseNameCount[baseName] = 0
        for time, val in zip(avgt, avgp):
            obsBaseNameCount[baseName] += 1

            obs = deepcopy(obsDefault)
            obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
            obs.OBSVAL = val * vFactor
            # additional for model result extraction
            obs._block_ = b
            obs._mtime_ = time * tFactor - offsetTime
            obses.append(obs)
            # output data file in original unit (instead of PEST/TOUGH2)
            # fo.write('%e %e\n' % (time, val))

        # generate batch plot entries
        plot = pressure_plot(
            b, baseName, [t-offsetTime/(60.0*60.0*24.0*365.25) for t in avgt], avgp,
            b+': '+','.join(ws))
        if PLOT_RAW_FIELD_DATA:
            for w in ws:
                ts, vs = p_bywell[w]['times'], p_bywell[w]['pressures']
                plot = plot_append_raw_data(plot, 'raw_'+w, ts, vs, xunit="year", yunit="bar")
        userEntry.batch_plot_entry.append(plot)

        # fo.close()
    return obses

def pressure_block_average_json_modelresult(geo,dat,lst,userEntry):
    obses = pressure_block_average_json_fielddata(geo, dat, userEntry)
    bs, tss = [], []
    for obs in obses:
        b = obs._block_
        t = obs._mtime_
        # print b, t
        if b not in bs:
            bs.append(b)
            tss.append([])
        tss[-1].append(t)

    tbls = lst.history([('e',b,FIELD['pres']) for b in bs])
    if tbls is None:
        raise Exception("Extraction Pressure history of listing results went wrong, blocks: %s" % ','.join(bs))
    # PyTOUGH checks result length, then just return the first set (hence a
    # tuple) if only one in the list, instead of a list
    if not isinstance(tbls, list):
        tbls = [tbls]

    from numpy import interp
    all_obs_vals = []
    for i, (b,ts) in enumerate(zip(bs,tss)):
        alltimes = tbls[i][0] # assuming all times are the same
        allpress = tbls[i][1]
        all_obs_vals += list(interp(ts,alltimes,allpress))

    return all_obs_vals

def private_history_data_with_boiling(userEntry):
    """ A special version of private_history_data() with enthalpy below boiling
    filtered out using user's '_BOILING_ABOVE_ENTH' value.
    """
    # default of 273 degree boiling, J/kg, unit same as internal object unit
    # (same as TOUGH2 unit)
    eboil = 1200.0e3
    if hasattr(userEntry.obsDefault, '_BOILING_ABOVE_ENTH'):
        eboil = userEntry.obsDefault._BOILING_ABOVE_ENTH

    entries = private_history_data(userEntry, 1000.0, 365.25*24.*60.*60.)
    filtered_entries = []
    for o, t in entries:
        if o.OBSVAL >= eboil:
            filtered_entries.append((o,t))
    return filtered_entries

def private_all_blocks_in_geners(name, all_gener_keys):
    """ return a list (unique and sorted) of block names from generators that
    matches name, with regular expression supported.  all_gener_keys should be a
    list of gener keys (tuple of block_name and gener_name).
    """
    import re
    from mulgrids import unfix_blockname,fix_blockname
    pattern = re.compile(name)
    bs = [b for (b,g) in all_gener_keys if pattern.match(unfix_blockname(g)) or pattern.match(g)]
    return sorted(list(set(bs)))

def private_boiling_plot(blockname, gener, datafile, timelist, obsname):
    """ for (timgui) batch plotting """
    return {
        "series": [
            {
                # "type": "HistoryPressureToBoil",
                "type": "HistoryPressureToBoilCO2",
                "block": blockname
            },
            {
                "type": "FrozenDataSeries",
                "name": obsname,
                "original_series": "",
                "frozen_x": list(timelist),
                "frozen_y": [0.0] * len(timelist),
                "xunit": "year",
                "yunit": "pascal",
            }
        ],
        "ylabel": "Pressure Difference to Boil (P-Psat)",
        "xlabel": "Time",
        "title": "%s - %s - %s" % (blockname, gener, datafile)
    }

def boiling_fielddata(geo, dat, userEntry):
    """ called by goPESTobs.py """
    from copy import deepcopy
    from gopest.common import private_cleanup_name
    obsDefault = userEntry.obsDefault

    entries = private_history_data_with_boiling(userEntry)
    if len(entries) == 0:
        raise Exception("User entry yields no observation: " + str(userEntry))

    psat_obses = []
    name = eval(userEntry.obsInfo[0])
    bs = private_all_blocks_in_geners(name, dat.generator.keys())
    if len(bs) == 0:
        msg = 'No GENERs matches with ' + name
        raise Exception(msg + "User entry yields no observation: " + str(userEntry))
    for b in bs:
        baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(b)
        if baseName not in obsBaseNameCount:
            obsBaseNameCount[baseName] = 0
        for o, t in entries:
            obsBaseNameCount[baseName] += 1
            obs = deepcopy(obsDefault)
            obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
            obs.OBSVAL = 0.0
            psat_obses.append(obs)
        # generate batch plot entries
        userEntry.batch_plot_entry.append(private_boiling_plot(
            b, name, userEntry.obsInfo[1], list(zip(*entries)[1])))
    return psat_obses

def boiling_modelresult(geo, dat, lst, userEntry):
    import numpy as np
    from t2thermo import sat
    from numpy import interp
    obsDefault = userEntry.obsDefault

    entries = private_history_data_with_boiling(userEntry)
    obses, timelist = map(list, zip(*entries))

    name = eval(userEntry.obsInfo[0])
    bs = private_all_blocks_in_geners(name, lst.generation.row_name)
    if len(bs) == 0:
        msg = 'No GENERs matches with ' + name
        raise Exception(msg + "User entry yields no observation: " + str(userEntry))
    selection = []
    for b in bs:
        # print "Boiling Block '%s' from '%s'" % (b, name)
        selection.append(('e',b,FIELD['temp']))
        selection.append(('e',b,FIELD['pres']))
    tbl = lst.history(selection)
    alltimes = tbl[0][0] # assuming all times are the same
    allpdiffs = []
    for i in range(len(bs)):
        ts = tbl[i*2][1]
        ps = tbl[i*2+1][1]
        pdiff_to_boil = []
        for (t,p) in zip(ts,ps):
            pdiff_to_boil.append(p - sat(t))
        pdiffs = interp(timelist,alltimes,np.array(pdiff_to_boil))
        allpdiffs += list(pdiffs)
    return allpdiffs

def private_enthalpy_plot(gname, datafile, timelist, enlist):
    """ for (timgui) batch plotting """
    return {
        "series": [
            {
                "type": "HistoryGenerEnthalpy",
                "gener": gname
            },
            {
                "type": "FrozenDataSeries",
                "name": datafile,
                "original_series": "",
                "frozen_x": [t/(365.25*24.*60.*60.) for t in timelist],
                "frozen_y": enlist,
                "xunit": "t2year",
                "yunit": "kJ/kg", #edited to kj/kg
            }
        ],
        "ylabel": "Average Enthalpy",
        "xlabel": "Time",
        "title": gname
    }

def enthalpy_fielddata(geo,dat,userEntry):
    """ called by goPESTobs.py """
    entries = private_history_data(userEntry, 1000.0, 365.25*24.*60.*60.)
    obses, times = map(list, zip(*entries))
    userEntry.batch_plot_entry.append(private_enthalpy_plot(
        eval(userEntry.obsInfo[0]), userEntry.obsInfo[1], times, [o.OBSVAL for o in obses]))
    return obses
def enthalpy_modelresult(geo,dat,lst,userEntry):
    # name,timelist
    name = eval(userEntry.obsInfo[0])
    entries = private_history_data(userEntry, 1000.0, 365.25*24.*60.*60.)
    obses, timelist = map(list, zip(*entries))
    """
    # nearest value
    import numpy as np
    lst.index = np.abs(lst.fulltimes-time).argmin()
    rows = lst.generation.rows_matching(('',name))
    total_heat, total_mass = 0.0, 0.0
    for r in rows:
        total_mass += r['Generation rate']
        total_heat += r['Generation rate'] * r['Enthalpy']
    if total_mass > 0.0:
        return total_heat / total_mass
    else:
        return 0.0
    """
    import numpy as np
    from mulgrids import unfix_blockname,fix_blockname
    allgs = lst.generation.row_name
    import re
    pattern = re.compile(name)
    gs = [(b,g) for (b,g) in allgs if pattern.match(unfix_blockname(g)) or pattern.match(g)]
    if len(gs) ==0:
        print('Warning, no GENERs matches with ', name)
        return 0.0
    # print "'%s' matches %i geners." % (name, len(gs))
    selection = []
    for gname in gs:
        selection.append(('g',gname,FIELD['rate']))
        selection.append(('g',gname,FIELD['enth']))
    tbl = lst.history(selection)
    alltimes = tbl[0][0] # assuming all times are the same
    total_heat = np.array([0.0 for i in range(len(alltimes))])
    total_mass = np.array([0.0 for i in range(len(alltimes))])
    for i in range(len(gs)):
        total_heat = total_heat + tbl[i*2+1][1] * tbl[i*2][1]
        total_mass = total_mass + tbl[i*2][1]
    average_enth = []
    for (mass,heat) in zip(total_mass,total_heat):
        if abs(mass) <= 1.0e-7:
            average_enth.append(0.0)
        else:
            average_enth.append(heat/mass)
    allenths = np.array(average_enth)
    from numpy import interp
    es = interp(timelist,alltimes,allenths)
    return list(es)

def gradient_by_central(xs, ys):
    import numpy as np
    if type(xs) is not np.ndarray:
        xs = np.array(xs)
    if type(ys) is not np.ndarray:
        ys = np.array(ys)
    z1 = np.hstack((ys[0], ys[:-1]))
    z2 = np.hstack((ys[1:], ys[-1]))

    dx1 = np.hstack((0.0, np.diff(xs)))
    dx2 = np.hstack((np.diff(xs), 0.0))

    d = (z2-z1) / (dx2+dx1)
    return d

def enthalpy_json_fielddata(geo,dat,userEntry):
    """ called by goPESTobs.py """
    jfilename = userEntry.obsInfo[0]
    customFilter = userEntry.customFilter
    offsetTime = userEntry.offsetTime
    obsDefault = userEntry.obsDefault
    vFactor = 1000.0 # assume given J/kg
    tFactor = 365.25*24.*60.*60. # assume given decimal years

    import numpy as np

    def weight_two_ends(def_weight, ends_factor, final_times):
        """ use this to increase or decrease weighting of start/end points in
        history  the middle of the data point will be kept as the set 'WEIGHT',
        then linearly  increases to the specified value here towards two ends.
        It can be either   larger than 1.0 or smaller than 1.0.
        """
        if (final_times[-1] - final_times[0]) < 1.0e-7:
            raise Exception("_ENDS_WEIGHT_FACTOR doesn't work if time range too small")
        final_weights = np.ones(len(final_times)) * def_weight
        half_t_range = (final_times[-1] - final_times[0]) / 2.0
        mid_t = (final_times[-1] + final_times[0]) / 2.0
        for i,t in enumerate(final_times):
            add_fac = abs(t - mid_t) / half_t_range * (ends_factor - 1.0)
            final_weights[i] = final_weights[i] * (add_fac + 1.0)
        return final_weights

    import json
    # 1st line is json file name (of all wells)
    with open(jfilename, 'r') as f:
        e_bywell = json.load(f)

    obses = []

    skipped_entryline = []
    skipped_gradient = []
    for oline in userEntry.obsInfo[1:]:
        wname = eval(oline)
        times, vals = [], []
        for time,val in zip(e_bywell[wname]['times'], e_bywell[wname]['enthalpy']):
            if eval(customFilter):
                times.append(time), vals.append(val)

        if len(times) == 0:
            raise Exception("User entry yields no observation: %s" % wname + str(userEntry))

        if hasattr(obsDefault, '_DESIRED_DATA_TIMES'):
            desired_times = obsDefault._DESIRED_DATA_TIMES
            final_times = [t for t in desired_times if times[0] <= t <= times[-1]]
            if hasattr(obsDefault, '_INTERP_LIMIT'):
                interp_limit = obsDefault._INTERP_LIMIT
                final_times = target_times(desired_times, interp_limit, list(zip(times,vals)))
            if len(final_times) == 0:
                raise Exception("User entry yields no observation: " + str(userEntry))
            from numpy import interp
            final_vals = list(interp(final_times,times,vals))
        else:
            final_times, final_vals = times, vals

        # set this to increase or decrease weighting of start/end points in
        # history the middle of the data point will be kept as the set 'WEIGHT',
        # then linearly increases to the specified value here towards two ends.
        # It can be either larger than 1.0 or smaller than 1.0.
        if hasattr(obsDefault, '_ENDS_WEIGHT_FACTOR'):
            ends_factor = float(obsDefault._ENDS_WEIGHT_FACTOR)
            final_weights = weight_two_ends(obsDefault.WEIGHT, ends_factor, final_times)
        else:
            final_weights = np.ones(len(final_times)) * obsDefault.WEIGHT

        from gopest.common import private_cleanup_name
        baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(wname)[:5]
        if baseName not in obsBaseNameCount:
            obsBaseNameCount[baseName] = 0

        for time, val, w in zip(final_times, final_vals, final_weights):
            obsBaseNameCount[baseName] += 1
            from copy import deepcopy
            obs = deepcopy(obsDefault)
            from gopest.common import private_cleanup_name
            obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
            obs.OBSVAL = val * vFactor
            obs.WEIGHT = w
            obses.append(obs)

        gpattern = wname
        if hasattr(obsDefault, '_WELL_TO_GENERS'):
            import json
            with open(obsDefault._WELL_TO_GENERS, 'r') as f:
                well_to_geners_dict = json.load(f)
            gpattern = well_to_geners_dict[wname]

        # generate batch plot entries
        plot = enthalpy_plot(
            gpattern, baseName, final_times, final_vals,
            ('Well %s' % wname) )
        if PLOT_RAW_FIELD_DATA:
            ts, vs = e_bywell[wname]['times'], e_bywell[wname]['enthalpy']
            plot = plot_append_raw_data(plot, 'raw_'+wname, ts, vs, xunit="year", yunit="kJ/kg") #edited to kJ/kg
        userEntry.batch_plot_entry.append(plot)

        if hasattr(obsDefault, '_GRADIENT_WEIGHT_FACTOR'):

            w = float(obsDefault._GRADIENT_WEIGHT_FACTOR) * obsDefault.WEIGHT
            if len(final_times) <= 1:
                print('enthalpy_json: User entry has too few data for gradient, skipping: %s' % wname)
                skipped_gradient.append(wname)
                continue
            final_gradients = gradient_by_central(final_times, final_vals)

            baseName = 'g' + obsDefault.OBSNME +'_'+ private_cleanup_name(wname)[:5]
            if baseName not in obsBaseNameCount:
                obsBaseNameCount[baseName] = 0

            for time, grad in zip(final_times, final_gradients):
                obsBaseNameCount[baseName] += 1
                from copy import deepcopy
                obs = deepcopy(obsDefault)
                from gopest.common import private_cleanup_name
                obs.OBGNME = obs.OBGNME + '_g'
                obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
                obs.OBSVAL = grad
                obs.WEIGHT = w
                obses.append(obs)
    # if skipped_gradient or skipped_entryline:
    #     print('See:\n' + str(userEntry))
    return obses

def private_remove_zeros(times, values):
    import numpy as np
    ts, vs = [],  []
    for t, v in zip(times, values):
        if v != 0.0:
            ts.append(t), vs.append(v)
    return np.array(ts), np.array(vs)

def enthalpy_json_modelresult(geo,dat,lst,userEntry):
    jfilename = userEntry.obsInfo[0]
    customFilter = userEntry.customFilter
    offsetTime = userEntry.offsetTime
    obsDefault = userEntry.obsDefault
    vFactor = 1.0 # assume given J/kg
    tFactor = 365.25*24.*60.*60. # assume given decimal years

    import json
    # 1st line is json file name (of all wells)
    with open(jfilename, 'r') as f:
        e_bywell = json.load(f)

    if hasattr(obsDefault, '_WELL_TO_GENERS'):
        with open(obsDefault._WELL_TO_GENERS, 'r') as f:
            well_to_geners_dict = json.load(f)

    alles = []

    for oline in userEntry.obsInfo[1:]:
        wname = eval(oline)
        times, vals = [], []
        for time,val in zip(e_bywell[wname]['times'], e_bywell[wname]['enthalpy']):
            if eval(customFilter):
                times.append(time), vals.append(val)

        if len(times) == 0:
            raise Exception("User entry yields no observation: " + str(userEntry))

        if hasattr(obsDefault, '_DESIRED_DATA_TIMES'):
            desired_times = obsDefault._DESIRED_DATA_TIMES
            final_times = [t for t in desired_times if times[0] <= t <= times[-1]]
            if hasattr(obsDefault, '_INTERP_LIMIT'):
                interp_limit = obsDefault._INTERP_LIMIT
                final_times = target_times(desired_times, interp_limit, list(zip(times,vals)))
            if len(final_times) == 0:
                raise Exception("User entry yields no observation: " + str(userEntry))
            from numpy import interp
            final_vals = list(interp(final_times,times,vals))
        else:
            final_times, final_vals = times, vals

        import numpy as np
        from mulgrids import unfix_blockname,fix_blockname
        allgs = lst.generation.row_name

        gpattern = wname
        if hasattr(obsDefault, '_WELL_TO_GENERS'):
            gpattern = well_to_geners_dict[wname]

        import re
        pattern = re.compile(gpattern)
        gs = [(b,g) for (b,g) in allgs if pattern.match(unfix_blockname(g)) or pattern.match(g)]
        if len(gs) ==0:
            print('Warning, no GENERs matches with ', gpattern)
            return 0.0
        # print "'%s' matches %i geners." % (wname, len(gs))
        selection = []
        for gname in gs:
            selection.append(('g',gname,FIELD['rate']))
            selection.append(('g',gname,FIELD['enth']))
        tbl = lst.history(selection)
        alltimes = tbl[0][0] # assuming all times are the same
        total_heat = np.array([0.0 for i in range(len(alltimes))])
        total_mass = np.array([0.0 for i in range(len(alltimes))])
        for i in range(len(gs)):
            total_heat = total_heat + tbl[i*2+1][1] * tbl[i*2][1]
            total_mass = total_mass + tbl[i*2][1]
        average_enth = []
        for (mass,heat) in zip(total_mass,total_heat):
            if abs(mass) <= 1.0e-7:
                average_enth.append(0.0)
            else:
                average_enth.append(heat/mass)
        allenths = np.array(average_enth)
        alltimes = alltimes / tFactor + offsetTime / tFactor
        if hasattr(obsDefault, '_REMOVE_ZEROS'):
            if obsDefault._REMOVE_ZEROS:
                alltimes, allenths = private_remove_zeros(alltimes, allenths)
        from numpy import interp
        # print "~~~~~", gpattern, len(final_times), len(alltimes), len(allenths)
        if len(alltimes) == 0:
            es = [0.0] * len(final_times)
            # es = [v - 300.0e3 for v in final_vals]
        else:
            # force enthalpy to be "reasonable", avoid crazy obj fn.
            allenths = [min(3.0e6,max(0.0,enth)) for enth in allenths]
            es = interp(final_times,alltimes,allenths)
        alles = alles + list(es)

        if hasattr(obsDefault, '_GRADIENT_WEIGHT_FACTOR'):
            if len(final_times) <= 1:
                # print('User entry has too few data for gradient, skipping: %s\n%s' % (wname, str(userEntry)))
                continue
            final_gradients = gradient_by_central(final_times, es)
            alles = alles + list(final_gradients)

    return alles

def enthalpy_plot(gname, dataname, ts, es, title):
    """ for (timgui) batch plotting """
    return {
        "series": [
            {
                "type": "HistoryGenerEnthalpy",
                "gener": gname
            },
            {
                "type": "FrozenDataSeries",
                "name": dataname,
                "original_series": "",
                "frozen_x": list(ts),
                "frozen_y": list(es),
                "xunit": "year",
                "yunit": "kJ/kg", #edited to kJ/kg
            }
        ],
        "ylabel": "Average Enthalpy",
        "xlabel": "Time",
        "title": title
    }


def boiling_json_fielddata(geo, dat, userEntry):
    """ called by goPESTobs.py """
    from copy import deepcopy
    from gopest.common import private_cleanup_name
    import json
    import numpy as np

    jfilename = userEntry.obsInfo[0]
    customFilter = userEntry.customFilter
    offsetTime = userEntry.offsetTime
    obsDefault = userEntry.obsDefault
    vFactor = 1.0 # assume given J/kg
    tFactor = 365.25*24.*60.*60. # assume given decimal years

    # default of 273 degree boiling, J/kg, unit same as internal object unit
    # (same as TOUGH2 unit)
    eboil = 1200.0e3
    if hasattr(userEntry.obsDefault, '_BOILING_ABOVE_ENTH'):
        eboil = userEntry.obsDefault._BOILING_ABOVE_ENTH

    # 1st line is json file name (of all wells)
    with open(jfilename, 'r') as f:
        e_bywell = json.load(f)

    obses = []
    boiling_blocks, blk_gener = [], {}

    for oline in userEntry.obsInfo[1:]:
        wname = eval(oline)
        gpattern = wname
        if hasattr(obsDefault, '_WELL_TO_GENERS'):
            import json
            with open(obsDefault._WELL_TO_GENERS, 'r') as f:
                well_to_geners_dict = json.load(f)
            gpattern = well_to_geners_dict[wname]

        times, vals = [], []
        for time,val in zip(e_bywell[wname]['times'], e_bywell[wname]['enthalpy']):
            if eval(customFilter):
                if val >= eboil:
                    times.append(time), vals.append(val)

        if len(times) == 0:
            print(wname, e_bywell[wname]['times'], e_bywell[wname]['enthalpy'])
            raise Exception("User entry has no boiling: " + wname + str(userEntry))
        if hasattr(obsDefault, '_DESIRED_DATA_TIMES'):
            desired_times = obsDefault._DESIRED_DATA_TIMES
            # print wname, desired_times, times, customFilter, e_bywell[wname]['times'], e_bywell[wname]['enthalpy']
            final_times = [t for t in desired_times if times[0] <= t <= times[-1]]
            if hasattr(obsDefault, '_INTERP_LIMIT'):
                interp_limit = obsDefault._INTERP_LIMIT
                final_times = target_times(desired_times, interp_limit, list(zip(times,vals)))
            final_vals = list(np.interp(final_times,times,vals))
        else:
            final_times, final_vals = times, vals
        if len(final_times) == 0:
            raise Exception("User entry yields no observation: " + wname + str(userEntry))

        if isinstance(dat, dict):
            # waiwera JSON input
            gen_keys = [(geo.block_name_list[geo.num_atmosphere_blocks:][gen['cell']], gen['name']) for gen in dat['source']]
        else:
            # aut2 dat
            gen_keys = dat.generator.keys()
        bs = private_all_blocks_in_geners(gpattern, gen_keys)
        if len(bs) == 0:
            print('gen_keys = ' + str(type(dat)) + ' ' + str(gen_keys))
            msg = "Check if dat.filename contains any matching geners:" + gpattern
            msg += "\n           ^^^^^ user entry yields no observation: " + wname + str(userEntry)
            raise Exception(msg)
        for b in bs:
            if b in boiling_blocks:
                msg1 = "block '%s' already used by previous gener '%s'" % (b, blk_gener[b])
                msg2 = "User entry yeilds additional boiling blocks: " + wname + str(userEntry)
                raise Exception('\n'.join([msg2, msg1]))
            blk_gener[b] = wname
            boiling_blocks.append(b)
            baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(b)
            # if baseName not in obsBaseNameCount:
            #     obsBaseNameCount[baseName] = 0
            for time, val in zip(final_times, final_vals):
                # obsBaseNameCount[baseName] += 1
                obs = deepcopy(obsDefault)
                # obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
                obs.OBSNME = unique_obs_name(obs.OBSNME, b)
                obs.OBSVAL = 0.0
                obs.WEIGHT = obsDefault.WEIGHT
                obs._block_ = b
                obs._mtime_ = time * tFactor - offsetTime
                obses.append(obs)
            # generate batch plot entries
            userEntry.batch_plot_entry.append(private_boiling_plot(
                b, gpattern, wname, final_times, baseName))

    return obses

def boiling_json_modelresult(geo,dat,lst,userEntry):
    from t2thermo import sat
    import numpy as np

    obses = boiling_json_fielddata(geo, dat, userEntry)
    bs, tss = [], []
    for obs in obses:
        b = obs._block_
        t = obs._mtime_
        # print b, t
        if b not in bs:
            bs.append(b)
            tss.append([])
        tss[-1].append(t)

    selection = []
    for b, ts in zip(bs, tss):
        selection.append(('e',b,FIELD['temp']))
        selection.append(('e',b,FIELD['pres']))
        selection.append(('e',b,FIELD['pco2']))

    allpdiffs = []
    tbl = lst.history(selection)
    alltimes = tbl[0][0] # assuming all times are the same
    for i,b in enumerate(bs):
        temps = tbl[i*3][1]
        press = tbl[i*3+1][1]
        pco2 = tbl[i*3+2][1]
        pdiff_to_boil = []
        for (t,p,p2) in zip(temps,press,pco2):
            pd = p - p2 - sat(t)
            if pd < 0.0:
                pd = 0.0 # allow super heating, treated as zero (good)
            pdiff_to_boil.append(pd)
        pdiffs = np.interp(tss[i], alltimes, np.array(pdiff_to_boil))
        allpdiffs += list(pdiffs)

    return allpdiffs


def private_history_data(userEntry,vFactor=1.0,tFactor=1.0):
    """ returns entries of history data read from files.

    If the default obs has (optional) property '_DESIRED_DATA_TIMES', the data
    from field data files will be interpolated into the desired times. The
    purpose of this was to make the observations more uniform in time, use with
    care.  It should be a list of time with the same unit and offset of the data
    files, NOT the rest of PEST/TOUGH2.
    """
    fieldDataFile = userEntry.obsInfo[1]
    customFilter = userEntry.customFilter
    offsetTime = userEntry.offsetTime
    obsDefault = userEntry.obsDefault
    f = open(fieldDataFile,'r')
    fo = open(fieldDataFile+'.obs', 'w')
    times, vals = [], []
    for line in f.readlines():
        if line.strip() == '': break
        time,val = [float(x) for x in line.split()[0:2]]
        if eval(customFilter):
            times.append(time), vals.append(val)

    if len(times) == 0:
        raise Exception("User entry yields no observation: " + str(userEntry))

    if hasattr(obsDefault, '_DESIRED_DATA_TIMES'):
        desired_times = obsDefault._DESIRED_DATA_TIMES
        final_times = [t for t in desired_times if times[0] <= t <= times[-1]]
        if hasattr(obsDefault, '_INTERP_LIMIT'):
            interp_limit = obsDefault._INTERP_LIMIT
            final_times = target_times(desired_times, interp_limit, list(zip(times,vals)))
        if len(final_times) == 0:
            raise Exception("User entry yields no observation: " + str(userEntry))
        from numpy import interp
        final_vals = list(interp(final_times,times,vals))
    else:
        final_times, final_vals = times, vals

    entries = []
    from gopest.common import private_cleanup_name
    baseName = obsDefault.OBSNME +'_'+ private_cleanup_name(userEntry.obsInfo[0])[:5]
    if baseName not in obsBaseNameCount:
        obsBaseNameCount[baseName] = 0
    for time, val in zip(final_times, final_vals):
        obsBaseNameCount[baseName] += 1

        from copy import deepcopy
        obs = deepcopy(obsDefault)
        from gopest.common import private_cleanup_name
        obs.OBSNME = baseName +'_'+ ('%04d' % obsBaseNameCount[baseName])
        obs.OBSVAL = val * vFactor
        entries.append((obs, time * tFactor - offsetTime))
        # output data file in original unit (instead of PEST/TOUGH2)
        fo.write('%e %e\n' % (time, val))
    fo.close()
    f.close()
    return entries


def private_has_re(text):
    """ simplified check of a string contains regular expression or not """
    special = '.^$*+?{}[]()'
    for s in special:
        if s in text:
            return True
    return False



import glob
import os
from shutil import copy2

from gopest.common import config

def current_iteration():
    its = [int(f.split('.')[-1]) for f in glob.glob('%s.jco.*' % config['pest']['case-name'])]
    its.append(0)
    return max(its)

def rename_latest_files(argv=[]):
    """ PEST_HP's file distribution will copy files from the best update slave
    to here.  They will be named as '*.999'.  This script renames these to the
    latest iteration number.
    """
    print('gopest save-iter-model (rename_latest_files.py):')

    # try:
    #     copy2('real_model.incon.999', 'real_model.incon')
    #     print("    real_model.incon.999 -> real_model.incon (copy)")
    # except Exception as e:
    #     print(" ", e)
    #     print("  failed to copy incon file for next iteration!")

    ii = current_iteration()
    print('  Current iteration is %i' % ii)

    for f in glob.glob('*.999'):
        newname = f.replace('.999', '.%i' % ii)
        print('    %s -> %s' % (f, newname))
        try:
            os.rename(f, newname)
        except Exception as e:
            print(e)



import os
import os.path
import string
import importlib.resources as resources
import shutil

# Access nested dictionary items via a list of keys
from functools import reduce  # forward compatibility for Python 3
import operator

import tomlkit

def check_required(finput, name, fdefault=None):
    if not os.path.isfile(finput):
        print("\nWarning! %s file '%s' is not found in current working directory:" % (name, finput))
        print("    %s\n" % os.getcwd())
        ans = input('Do you want goPEST to create a default file? (y/n) ')
        if 'y' in ans.lower():
            if fdefault is None:
                fdefault = finput
            default_finput = resources.files('gopest.data') / fdefault
            with resources.as_file(default_finput) as f:
                shutil.copyfile(f, finput)
                print("Default %s file '%s' created, please review and re-run " \
                      "command 'gopest init'." % (name, finput))
            exit(0)
        else:
            print("Error! File '%s' is required for goPEST to work." % finput)
            print('Existing...')
            exit(1)


check_required('goPESTconfig.toml', 'Configuration')

""" this allows gopest.common.config to be used directly, eg.
        from gopest.common import config as cfg
        print(cfg['pest']['executable'])
        print(cfg['simulator']['executable'])
"""
with open('goPESTconfig.toml', 'r') as f:
    config = tomlkit.load(f)

    if 'silent_slaves' not in config['pest']:
        config['pest']['silent_slaves'] = True

""" this allows gopest.common.runtime to be used directly, eg.
        from gopest.common import runtime
        print(runtime['filename']['fincon'])
        print(runtime['filename']['fdatns'])
"""
runtime = {}
def runtime_filenames(check=False):
    """ work out filenames for all internal model files """
    def getext(fn): return os.path.splitext(fn)[1].lower()
    all_geoms = ['g_real_model' + getext(f) for f in config['model']['original']['geometry-files']]
    input_typ = config['simulator']['input-type']
    output_typ = config['simulator']['output-type']
    sequence = config['model']['sequence']
    if input_typ == 'aut2':
        fsave = 'real_model_' + sequence[0] + '.save'
        fincon = 'real_model_' + 'incon' + '.incon'
        fdato = 'real_model_' + 'original' + '.dat'
        fdats = ['real_model_' + seq + '.dat' for seq in sequence]
        fincs = ['real_model_' + seq + '.incon' for seq in sequence]
        fsavs = ['real_model_' + seq + '.save' for seq in sequence]
        if output_typ == 'h5':
            flsts = ['real_model_' + seq + '.h5' for seq in sequence]
        else:
            flsts = ['real_model_' + seq + '.listing' for seq in sequence]
    elif input_typ == 'waiwera':
        fsave = 'real_model_' + sequence[0] + '.h5'
        fincon = 'real_model_' + 'incon' + '.h5'
        fdato = 'real_model_' + 'original' + '.json'
        fdats = ['real_model_' + seq + '.json' for seq in sequence]
        flsts = ['real_model_' + seq + '.h5' for seq in sequence]
        fincs = ['real_model_' + seq + '_inc.h5' for seq in sequence]
        fsavs = ['real_model_' + seq + '.h5' for seq in sequence]
    filenames = {
        # single string properties
        'geom': 'g_real_model.dat',
        'save': fsave,
        'incon': fincon,
        'dat_orig': fdato,
        # following peoperties are lists
        'all_geoms': all_geoms,
        'dat_seq': fdats,
        'lst_seq': flsts,
        'inc_seq': fincs,
        'sav_seq': fsavs,
    }
    if check:
        for k,fnames in filenames.items():
            if not isinstance(fnames, list):
                fnames = [fnames]
            for fn in fnames:
                if not os.path.exists(fn):
                    raise Exception('Cannot find file %s in the working dir.' % fn)
    return filenames
runtime['filename'] = runtime_filenames()

########## utility classes and functions
class TwoWayDict(dict):
    def __init__(self,copyFrom=None):
        if copyFrom is not None:
            for k in copyFrom.keys():
                self[k] = copyFrom[k]
    def __len__(self):
        return dict.__len__(self) / 2
    def __setitem__(self, key, value):
        if key in self: raise Exception('TwoWayDict repeated key: %s' % str(key))
        dict.__setitem__(self, key, value)
        if value in self: raise Exception('TwoWayDict repeated key: %s' % str(value))
        dict.__setitem__(self, value, key)

class Singleton(object):
    __single = None # the one, true Singleton
    def __new__(classtype, *args, **kwargs):
        # Check to see if a __single exists already for this class
        # Compare class types instead of just looking for None so
        # that subclasses will create their own __single objects
        if classtype != type(classtype.__single):
            classtype.__single = object.__new__(classtype, *args, **kwargs)
        return classtype.__single

def readList(f):
    """ reads a file and returns a list of entries, an entry is a list strings
        all lines starts with '#' are ignored """
    entryName, entries = [], []
    for line in f.readlines():
        if len(line.strip()) == 0: continue  # empty lines
        if line.strip()[0] == '#': continue  # comment lines
        if line.strip()[0] == '[':           # [something] item opening
            tmp = line.strip()
            ikeyend=tmp.find(']')
            keyword=tmp[1:ikeyend]
            if keyword=='END': break         # [END] finishes all
            entryName.append(keyword)
            entries.append([])
            continue
        # all other non-empty lines parts of last entry (until [END])
        # error if some non-comment line apear above first keyword
        entries[-1].append(line)
    return (entryName, entries)

def updateDict(dic,lines):
    """ evaluate lines and update the content of the argument dictionary.  A KeyError
        exception is raised if the key from lines is not in dic.  Each line in lines
        is a simple string like 'key = val'. """
    keys,vals = [],[]
    for line in lines:
        if line.count('=') != 1: raise Exception
        k,v = line.split('=')
        if k.strip() not in dic:
            print(dic.keys())
            raise KeyError
        dic[k.strip()] = eval(v)
    return dic

def merge_dols(dol1, dol2):
    """ merging dicts of lists into a new dict of lists. """
    keys = set(dol1).union(dol2)
    no = []
    return dict((k, dol1.get(k, no) + dol2.get(k, no)) for k in keys)

def updateObj(obj,lines):
    """ evaluate lines and modify the members of the argument object, A KeyError
        exception is raised if the key from lines is not a member of the object.
        Each line in lines is a simple string like 'key = val'. """
    keys,vals = [],[]
    for line in lines:
        if line.count('=') != 1: raise Exception
        k,v = line.split('=')
        # if k.strip() not in obj.__dict__.keys():
        #     print(k, ' not supported, try: ', obj.__dict__.keys())
        #     raise KeyError
        setattr(obj,k.strip(),eval(v))
    return obj

def private_cleanup_name(s):
    """ cleans up string so no space, no funny symbols.  it will remove all
    punctuations chars apart from those specified in rep, which will be '_' """
    rep = ' .*'
    rmv = string.punctuation
    for c in rep: rmv = rmv.replace(c,'_')
    for c in rmv: s = s.replace(c,'')
    for c in rep: s = s.replace(c,'_')
    return s

def getFromDict(dataDict, mapList):
    return reduce(operator.getitem, mapList, dataDict)

def setInDict(dataDict, mapList, value):
    getFromDict(dataDict, mapList[:-1])[mapList[-1]] = value




import time
import json
import inspect

from mulgrids import *
from t2data import *
from t2listing import *

from gopest.common import Singleton
from gopest.common import TwoWayDict
from gopest.common import readList
from gopest.common import updateObj
from gopest.common import merge_dols
from gopest import obs_def

from gopest.utils.waiwera_listing import wlisting
from gopest.utils.t2listingh5 import t2listingh5

OBS_USER_FUNC = dict(inspect.getmembers(obs_def,inspect.isfunction))
OBS_ALIAS = TwoWayDict(obs_def.shortNames)

class PestObsDataName(Singleton):
    """ remembers a list of observation data and observation data """
    def __init__(self):
        self.basenames = set([])
    def newName(self,basename):
        apnd = ' 01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
        i = 0
        newname = (basename + apnd[i]).strip()
        while newname in self.basenames:
            i += 1
            if i >= len(apnd):
                raise Exception('Unable to find unused obseration name for:' + basename)
                break
            newname = (basename + apnd[i]).strip()
        return newname
    def add(self,newname):
        self.basenames.add(newname)

class PestObservData(object):
    def __init__(self,OBSNME='',OBSVAL=0.0,WEIGHT=1.0,OBGNME=''):
        self.OBSNME=OBSNME
        self.OBSVAL=OBSVAL
        self.WEIGHT=WEIGHT
        self.OBGNME=OBGNME
    def __repr__(self):
        return  self.OBSNME +' '+ str(self.OBSVAL) +' '+ str(self.WEIGHT) +' '+ self.OBGNME

class UserEntryObserv(object):
    """
        user defined routines
    """
    def __init__(self,obsType,obsInfo,fieldDataFile,customFilter,
        offsetTime,obsDefault=PestObservData()):
        self.obsType = obsType # a string
        self.obsInfo = obsInfo # a list of anything
        self.fieldDataFile = fieldDataFile # a string
        self.customFilter = customFilter # an expression string
        self.offsetTime = offsetTime # a number to offset model time
        from copy import deepcopy
        self.obsDefault = deepcopy(obsDefault)

        self.all_obses = None
        self.all_pst_lines = None # should be a list of strings
        self.all_ins_lines = None # should be a list of strings
        self.all_obf_lines = None # should be a list of strings

        # this is expected to be directly written into self, by custom function
        self.batch_plot_entry = []
        self.coverage = {} # keyed by OBGNME, each group is a list of model blocks with data
    def __repr__(self):
        return '\n'.join([
            '','[Obs]',
            self.obsType,
            ';'.join([str(info) for info in self.obsInfo]),
            self.fieldDataFile,
            self.customFilter,
            str(self.obsDefault),
            ''])
    def makeObsDataInsLines(self,geo,dat):
        """ generate a tuple of two lists:
            - list of Pest * observation data """
        # maybe + ('%5s' % str(self.obsInfo[0])) with fill?
        if self.obsDefault.OBSNME == '':
            newBaseName = PestObsDataName().newName(OBS_ALIAS[self.obsType])
            PestObsDataName().add(newBaseName)
            # user def routines use this as basename
            self.obsDefault.OBSNME = newBaseName
        # use obsType if not specified by user
        if self.obsDefault.OBGNME == '':
            self.obsDefault.OBGNME = self.obsType

        ### this line does all the work
        self.all_obses = OBS_USER_FUNC[self.obsType+'_fielddata'](geo,dat,self)

        self.all_pst_lines = []
        self.all_ins_lines = []
        for obs in self.all_obses:
            self.all_ins_lines.append('l1 [%s]21:41' % obs.OBSNME)
            self.all_pst_lines.append(' %-20s %20.13e %12.5e %s' % (obs.OBSNME,
                obs.OBSVAL,obs.WEIGHT,obs.OBGNME))

    def makeObfLines(self,geo,dat,lst):
        """ generate a list of:
            - values for obf file that PEST requires """

        ### this line does all the work
        obfValues = OBS_USER_FUNC[self.obsType+'_modelresult'](geo,dat,lst,self)
        self.all_obf_lines = []
        for (v,obs) in zip(obfValues,self.all_obses):
            self.all_obf_lines.append('%-20s %20.13e' % (obs.OBSNME,v))

def readUserObservation(userListName):
    """ returns a list of UserEntryObserv from reading the file with name
        userObsListName """
    userEntries = []
    f = open(userListName,'r')
    entryName, entry = readList(f)
    f.close()
    # some defaults even if no sections exists:
    obsDefault = PestObservData()
    customFilter = 'True'
    offsetTime = 0.0
    for i,en in enumerate(entryName):
        if en == 'ObservationType':
            obsType = entry[i][0].strip()
            continue
        if en == 'DataFilter':
            if len(entry[i]) == 0:
                # reset
                customFilter = 'True'
            else:
                customFilter = entry[i][0]
        if en == 'DataTimeOffset':
            if len(entry[i]) == 0:
                # reset
                offsetTime = 0.0
            else:
                offsetTime = float(eval(entry[i][0]))
        if en == 'Defaults':
            obsDefault = updateObj(obsDefault,entry[i])
        # if en == 'Obs':
        #     if len(entry[i]) < 2: raise Exception
        #     if ',' in entry[i][0]:
        #         obsInfo = list(eval(entry[i][0]))
        #     else:
        #         obsInfo = [eval(entry[i][0])]
        #     for fieldDataFile in entry[i][1:]:
        #         userEntries.append(UserEntryObserv(obsType,obsInfo,
        #             fieldDataFile.strip(),customFilter,offsetTime,
        #             obsDefault))
        if en == 'Obs':
            if len(entry[i]) < 1:
                raise Exception("An empty [Obs] entry is found in goPESTobs.list")
            # pass the list of lines in, user functions to deal with them
            obsInfo = [s.rstrip('\n') for s in entry[i]]
            fieldDataFile = ''
            userEntries.append(UserEntryObserv(obsType,obsInfo,
                fieldDataFile.strip(),customFilter,offsetTime,
                obsDefault))
    return userEntries

def generate_obses_and_ins(fgeo, fdat, insToWrite, fobses, fplts='goPESTobs.json', fcovs='goPESTobs.coverage'):
    """ reads goPESTobs.list and generate observation data lines and instruction
    file for PEST """
    # reset unique obs name
    obs_def.obsBaseNameCount = {}
    geo = mulgrid(fgeo)
    if fdat.endswith('.json'):
        with open(fdat, 'r') as f:
            dat = json.load(f)
    else:
        dat = t2data(fdat)

    userEntries = readUserObservation('goPESTobs.list')
    pstLines, insLines, plots, coverage = [], [], [], {}
    for ue in userEntries:
        ue.makeObsDataInsLines(geo,dat)
        pstLines = pstLines + ue.all_pst_lines
        insLines = insLines + ue.all_ins_lines
        plots += ue.batch_plot_entry
        coverage = merge_dols(coverage, ue.coverage)

    f = open(insToWrite, 'w')
    f.write('pif #\n')
    for line in insLines:
        f.write(line+'\n')
    f.close()

    obs = open(fobses, 'w')
    for line in pstLines:
        obs.write(line + '\n')
    obs.close()

    plt = open(fplts, 'w')
    json.dump(plots, plt, indent=4, sort_keys=True)
    plt.close()

    cov = open(fcovs, 'w')
    json.dump(coverage, cov, indent=4, sort_keys=True)
    cov.close()

def read_from_real_model(fgeo, fdat, flst, fobf, waiwera=False):
    """ This reads TOUGH2's results and write in appropriate format into obf
    file for PEST """
    # reset unique obs name
    obs_def.obsBaseNameCount = {}
    geo = mulgrid(fgeo)
    if waiwera:
        with open(fdat, 'r') as f:
            dat = json.load(f)
        lst = wlisting(flst, geo, fjson=fdat)
    else:
        dat = t2data(fdat)
        if flst.lower().endswith('.h5'):
            lst = t2listingh5(flst)
        else:
            lst = t2listing(flst)

    userEntries = readUserObservation('goPESTobs.list')
    obfLines = []
    for ue in userEntries:
        ue.makeObsDataInsLines(geo,dat)
        ue.makeObfLines(geo,dat,lst)
        obfLines = obfLines + ue.all_obf_lines

    if flst.lower().endswith('.listing'):
        lst.close()

    f = open(fobf,'w')
    for line in obfLines:
        f.write(line+'\n')
    f.close()

def goPESTobs(argv=[]):
    START_TIME = time.time()

    userlistname = 'goPESTobs.list'

    if len(argv) not in [4,5]:
        print('to generate PEST .pst observation section and .ins: ')
        print('     gopest obs geo dat newPESTins')
        print('to read Tough2 results and write result file for PEST to read:')
        print('     gopest obs geo dat lst newPESTobf')

    if len(argv) == 4:
        fgeo = argv[1]
        fdat = argv[2]
        insToWrite = argv[3]

        fobses = 'pest_obs_data'
        fplts = 'goPESTobs.json'
        fcovs = 'goPESTobs.coverage'
        generate_obses_and_ins(fgeo, fdat, insToWrite,
                               fobses, fplts, fcovs)

    if len(argv) == 5:
        
        fgeo = argv[1]
        fdat = argv[2]
        flst = argv[3]
        obfToWrite = argv[4]

        geo = mulgrid(fgeo)
        if fdat.lower().endswith('.json'):
            with open(fdat, 'r') as f:
                dat = json.load(f)
        else:
            dat = t2data(fdat)
        
        if flst.lower().endswith('.listing'):
            lst = t2listing(flst)
        elif flst.lower().endswith('.h5'):
            if fdat.lower().endswith('.json'):
                lst = wlisting(flst, geo, fjson=fdat)
            else:
                lst = t2listingh5(flst)



        userEntries = readUserObservation(userlistname)
        obfLines = []
        for ue in userEntries:
            ue.makeObsDataInsLines(geo,dat)
            ue.makeObfLines(geo,dat,lst)
            obfLines = obfLines + ue.all_obf_lines


        f = open(obfToWrite,'w')
        for line in obfLines:
            f.write(line+'\n')
        f.close()
        
    # print('goPESTobs finished after', (time.time() - START_TIME), 'seconds')
        





"""
Copyright 2013, 2014 University of Auckland.

This file is part of TIM (Tim Isn't Mulgraph).

    TIM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    TIM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TIM.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
Wrap AUTOUGH2's hdf5 output as t2listing
"""

import h5py
import numpy as np

from t2listing import *
# from t2listing import listingtable
from mulgrids import fix_blockname, unfix_blockname

from pprint import pprint as pp
import unittest

class h5table(listingtable):
    """ Class emulating the listingtable class in PyTOUGH.

    Class for table in listing file, with values addressable by index (0-based)
    or row name, and column name: e.g. table[i] returns the ith row (as a
    dictionary), table[rowname] returns the row with the specified name, and
    table[colname] returns the column with the specified name.

    !!! IMPORTANT !!!
    .index needs to be set whenever listing object changed time index
    """
    def __init__(self, cols, rows, h5_table,
                 num_keys = 1, allow_reverse_keys = False,
                 index = 0):
        """ The row_format parameter is a dictionary with three keys,
        'key','index' and 'values'.  These contain the positions, in each row of
        the table, of the start of the keys, index and data fields.  The
        row_line parameter is a list containing, for each row of the table, the
        number of lines before it in the listing file, from the start of the
        table.  This is needed for TOUGH2_MP listing files, in which the rows
        are not in index order and can also be duplicated.

        h5_table should be the table within the h5 file.
        """
        self.column_name = cols
        self.row_name = rows
        self.num_keys = num_keys
        self.allow_reverse_keys = allow_reverse_keys
        self._col = dict([(c,i) for i,c in enumerate(cols)])
        self._row = dict([(r,i) for i,r in enumerate(rows)])
        self._h5_table = h5_table
        self._index = index # time index

    def __repr__(self):
        # h5 table lst._h5['element'][time index, eleme index, field index]
        return repr(self.column_name) + '\n' + repr(self._h5_table[self._index, :, :])

    def __getitem__(self, key):
        if isinstance(key, int):
            return dict(zip(['key'] + self.column_name, [self.row_name[key]] +
                            list(self._h5_table[self._index, key, :])))
        else:
            if key in self.column_name:
                return self._h5_table[self._index, :, self._col[key]]
            elif key in self.row_name:
                rowindex = self._row[key]
                return dict(zip(['key'] + self.column_name,
                                [self.row_name[rowindex]] +
                                list(self._h5_table[self._index, rowindex, :])))
            elif len(key) > 1 and self.allow_reverse_keys:
                revkey = key[::-1] # try reversed key for multi-key tables
                if revkey in self.row_name:
                    rowindex = self._row[revkey]
                    return dict(zip(['key'] + self.column_name,
                                    [self.row_name[rowindex][::-1]] +
                                    list(-self._h5_table[self._index, rowindex, :])))
            else: return None

    def __add__(self, other):
        raise NotImplementedError
        """Adds two listing tables together."""
        if self.column_name == other.column_name and self.row_name == other.row_name:
            from copy import copy
            result = listingtable(copy(self.column_name), copy(self.row_name), num_keys = self.num_keys,
                                  allow_reverse_keys = self.allow_reverse_keys)
            result._data = self._data + other._data
            return result
        else: raise Exception("Incompatible tables: can't be added together.")

    def __sub__(self, other):
        raise NotImplementedError
        """Subtracts one listing table from another."""
        if self.column_name == other.column_name and self.row_name == other.row_name:
            from copy import copy
            result = listingtable(copy(self.column_name), copy(self.row_name), num_keys = self.num_keys,
                                  allow_reverse_keys = self.allow_reverse_keys)
            result._data = self._data - other._data
            return result
        else: raise Exception("Incompatible tables: can't be subtracted.")


class t2listingh5(object):
    def __init__(self, filename):
        """
        """
        self._table = {}
        self._h5 = h5py.File(filename, 'r')
        self.filename = filename
        self.setup()
        self.simulator = 'AUTOUGH2_H5'

    def close(self):
        self._h5.close()

    def setup(self):
        self.fulltimes = self._h5['fulltimes']['TIME']
        self.num_fulltimes = len(self.fulltimes)
        self._index = 0 # this is the internal one
        ### element table
        if 'element' in self._h5:
            cols = [x.decode('utf-8') for x in self._h5['element_fields']]
            blocks = [fix_blockname(x.decode('utf-8')) for x in self._h5['element_names']]
            table = h5table(cols, blocks, self._h5['element'], num_keys=1)
            self._table['element'] = table
        ### connection table
        if 'connection' in self._h5:
            cols = [x.decode('utf-8') for x in self._h5['connection_fields'][:]]
            b1 = [fix_blockname(x.decode('utf-8')) for x in self._h5['connection_names1'][:]]
            b2 = [fix_blockname(x.decode('utf-8')) for x in self._h5['connection_names2'][:]]
            table = h5table(cols, list(zip(b1,b2)), self._h5['connection'], num_keys=2,
                            allow_reverse_keys=True)
            self._table['connection'] = table
        ### generation table
        if 'generation' in self._h5:
            cols = [x.decode('utf-8') for x in self._h5['generation_fields'][:]]
            blocks = [fix_blockname(x.decode('utf-8')) for x in self._h5['generation_eleme'][:]]
            geners = [fix_blockname(x.decode('utf-8')) for x in self._h5['generation_names'][:]]
            table = h5table(cols, list(zip(blocks,geners)), self._h5['generation'], num_keys=1)
            self._table['generation'] = table
        # makes tables in self._table accessible as attributes
        for key,table in self._table.items():
            setattr(self, key, table)
        # have to be get first table ready
        self.index = 0

    def history(self, selection, short=False, start_datetime=None):
        """
        short is not used at the moment
        """
        if isinstance(selection, tuple):
            selection = [selection]
        results = []
        for tbl,b,cname in selection:
            table_name, bi, fieldi = self.selection_index(tbl, b, cname)
            if bi < 0:
                bi = len(self.block_name_index) + bi
            ### important to convert cell index
            ys = self._h5[table_name][:,bi,fieldi]
            results.append((self.fulltimes, ys))
        if len(results) == 1: results = results[0]
        return results

    def selection_index(self, tbl, b, field):
        dname = {
            'e': 'element',
            'c': 'connection',
            'g': 'generation',
        }
        def eleme_index(b):
            if isinstance(b, str):
                bi = self.block_name_index[b]
            elif isinstance(b, int):
                bi = b
            else:
                raise Exception('.history() block must be an int or str: %s (%s)' % (str(b),str(type(b))))
            return bi
        def conne_index(b):
            if isinstance(b, tuple):
                bi = self.connection_name_index[(str(b[0]), str(b[1]))]
            elif isinstance(b, int):
                bi = b
            else:
                raise Exception('.history() conne must be an int or (str,str): %s (%s)' % (str(b),str(type(b))))
            return bi
        def gener_index(b):
            if isinstance(b, tuple):
                bi = self.generation_name_index[(str(b[0]), str(b[1]))]
            elif isinstance(b, int):
                bi = b
            else:
                raise Exception('.history() gener must be an int or (str,str): %s (%s)' % (str(b),str(type(b))))
            return bi
        iname = {
            'e': eleme_index,
            'c': conne_index,
            'g': gener_index,
        }
        if not hasattr(self, 'field_index'):
            self.field_index = {}
            for n,nn in dname.items():
                for i,ff in enumerate(self._h5[nn + '_fields']):
                    self.field_index[(n,ff)] = i
        return dname[tbl], iname[tbl](b), self.field_index[(tbl,field)]

    @property
    def block_name_index(self):
        if not hasattr(self, '_block_name_index'):
            self._block_name_index = {}
            # self._block_name_index.update({str(e):i for i,e in enumerate(self._h5['element_names'])})
            self._block_name_index.update({fix_blockname(str(e)):i for i,e in enumerate(self._h5['element_names'])})
        return self._block_name_index

    @property
    def connection_name_index(self):
        if not hasattr(self, '_connection_name_index'):
            a = self._h5['connection_names1']
            b = self._h5['connection_names2']
            self._connection_name_index = {}
            self._connection_name_index.update({(fix_blockname(str(x[0])),fix_blockname(str(x[1]))):i for i,x in enumerate(zip(a,b))})
        return self._connection_name_index

    @property
    def generation_name_index(self):
        if not hasattr(self, '_generation_name_index'):
            a = self._h5['generation_eleme']
            b = self._h5['generation_names']
            self._generation_name_index = {}
            # self._generation_name_index.update({(str(x[0]),str(x[1])):i for i,x in enumerate(zip(a,b))})
            # self._generation_name_index.update({(fix_blockname(str(x[0])),(str(x[1]))):i for i,x in enumerate(zip(a,b))})
            # self._generation_name_index.update({((str(x[0])),fix_blockname(str(x[1]))):i for i,x in enumerate(zip(a,b))})
            self._generation_name_index.update({(fix_blockname(str(x[0])),fix_blockname(str(x[1]))):i for i,x in enumerate(zip(a,b))})
        return self._generation_name_index


    def read_tables(self):
        """ copy values from h5 into listingtables, with slicing """
        if 'element' in self.table_names:
            self.element._index = self.index
            # for i,cname in enumerate(self.element.column_name):
            #     self.element._data[:,i] = self._h5['element'][self._index, :, i]
        if 'connection' in self.table_names:
            self.connection._index = self.index
            # for i,cname in enumerate(self.connection.column_name):
            #     self.connection._data[:,i] = self._h5['connection'][self._index, :, i]
        if 'generation' in self.table_names:
            self.generation._index = self.index
            # for i,cname in enumerate(self.generation.column_name):
            #     self.generation._data[:,i] = self._h5['generation'][self._index, :, i]

    def get_index(self): return self._index
    def set_index(self, i):
        self._index = i
        if self._index < 0: self._index += self.num_fulltimes
        self.read_tables()
    index = property(get_index, set_index)

    def first(self): self.index = 0
    def last(self): self.index = -1
    def next(self):
        """Find and read next set of results; returns false if at end of listing"""
        more = self.index < self.num_fulltimes - 1
        if more: self.index += 1
        return more
    def prev(self):
        """Find and read previous set of results; returns false if at start of listing"""
        more = self.index > 0
        if more: self.index -= 1
        return more

    def get_table_names(self):
        return sorted(self._table.keys())
    table_names = property(get_table_names)

    def get_time(self): return self.fulltimes[self.index]
    def set_time(self, t):
        if t < self.fulltimes[0]: self.index = 0
        elif t > self.fulltimes[-1]: self.index = -1
        else:
            dt = np.abs(self.fulltimes - t)
            self.index = np.argmin(dt)
    time = property(get_time, set_time)


class test_fivespot(unittest.TestCase):
    def setUp(self):
        self.lst_h = t2listingh5('fivespot.h5')
        self.lst_t= t2listing('expected.listing')

    def test_match_tables(self):
        # check row and column names
        def check_names(tbl):
            tbl_h = getattr(self.lst_h, tbl)
            tbl_t = getattr(self.lst_t, tbl)
            self.assertEqual(tbl_h.row_name, tbl_t.row_name)
            for i,field in enumerate(tbl_h.column_name):
                if tbl_t.column_name[i] in field:
                    match = True
                else:
                    match = False
                self.assertEqual(match, True, '%s: column name mismatch' % tbl)
        for tbl in ['element', 'connection', 'generation']:
            check_names(tbl)
        # check table values, after change index also
        def check_tables():
            rtol = 1.0e-5 # roughly 4~5 significant digits from text listing file
            for field in ['Temperature', 'Pressure', 'Vapour saturation']:
                np.testing.assert_allclose(self.lst_h.element[field],
                                           self.lst_t.element[field], rtol=rtol)
            for field in ['Mass flow', 'Enthalpy', 'Heat flow']:
                np.testing.assert_allclose(self.lst_h.connection[field],
                                           self.lst_t.connection[field], rtol=rtol)
            for field in ['Generation rate', 'Enthalpy']:
                np.testing.assert_allclose(self.lst_h.generation[field],
                                           self.lst_t.generation[field], rtol=rtol)
        check_tables()
        self.lst_h.last(); self.lst_t.last()
        check_tables()
        self.lst_h.first(); self.lst_t.first()
        check_tables()

        # check table with element index
        def check_table_by_index(i):
            tbl_h = self.lst_h.element[i]
            tbl_t = self.lst_t.element[i]
            for k in tbl_h.keys():
                if k == 'key':
                    self.assertEqual(tbl_h[k], tbl_t[k])
                else:
                    np.testing.assert_approx_equal(tbl_h[k], tbl_t[k], significant=5)
        for i in range(len(self.lst_h.element.row_name)):
            check_table_by_index(i)

        # check table with element name
        def check_table_by_name(b):
            tbl_h = self.lst_h.element[b]
            tbl_t = self.lst_t.element[b]
            for k in tbl_h.keys():
                if k == 'key':
                    self.assertEqual(tbl_h[k], tbl_t[k])
                else:
                    np.testing.assert_approx_equal(tbl_h[k], tbl_t[k], significant=5)
        for b in self.lst_h.element.row_name:
            check_table_by_index(b)

    def test_match_history(self):
        self.assertEqual(self.lst_h.num_fulltimes, self.lst_t.num_fulltimes)
        np.testing.assert_allclose(self.lst_h.fulltimes, self.lst_t.fulltimes)
        rtol = 1.0e-5
        # seems allfield names are identical with fivespot's EOS
        for sel in [('e', 'AA106', 'Pressure'),
                    ('e', 'AA 66', 'Temperature'),
                    ('c', ('AA 66', 'AA 67'), 'Mass flow'),
                    ('g', ('AA 11', 'PRO 1'), 'Generation rate'),
                    ('g', ('AA 11', 'PRO 1'), 'Enthalpy'),
                    ]:
            xs_h, ys_h = self.lst_h.history(sel)
            xs_t, ys_t = self.lst_t.history(sel)
            np.testing.assert_allclose(xs_h, xs_t, rtol=rtol)
            np.testing.assert_allclose(ys_h, ys_t, rtol=rtol)


if __name__ == '__main__':
    unittest.main(verbosity=2)



import os
import sys
import shutil
import time
from subprocess import Popen, DEVNULL
from multiprocessing import cpu_count

from gopest.common import config as cfg
from gopest.common import runtime

NUM_SLAVES = cfg['pest']['num_slaves']
SILENT_SLAVES = cfg['pest']['silent_slaves']
PST_NAME = cfg['pest']['case-name']
PORT = cfg['pest']['port']
SWITCHES = " ".join(cfg['pest']['switches'])

TOUGH2 = cfg['simulator']['executable']
BEOPEST = cfg['pest']['executable']
PESTDIR = cfg['pest']['dir']

"""
This script helps to launch BeoPEST on the case in current directory (either
'case.pst' or 'case_svda.pst').  Use:

    python run_beopest.py

Requires:
    - Python 2.6 or 2.7, NumPy
    - PEST and TOUGH2 executables, see settings above.

Most of the important settings are listed above.  Instead of letting PEST runs
TOUGH2 models directly, actual model runs were controlled by the pest_model.py
script, which uses the goPEST interface and run_ns_pr.py.  (goPEST is a little
like PLPROC)

In the actual workflow, there are only a couple files that need attention:

    - goPESTpar.list    a list of parameters (goPESTpar.py translates this into
        actual PEST parameter data)
    - goPESTobs.list    a list of observations (goPESTobs.py translates this
        into actual PEST observation data)

Brief workflow:

    1. edit goPESTpar.list and goPESTobs.list

    2. run 'python make_case_pst.py', which processed two lists above and
    updates case.pst.  This also initialise the parameter values from the values
    in the real_model_original.dat (.pdat).

    3. inspect/modify/check case.pst

    4. run 'python run_beopest.py'

    5. to obtain the model with optimised parameters run 'python
    make_real_model.py' or 'python make_real_model_svda.py', now
    real_model.dat is the updated model, run_ns_pr.py can be used to run    this
    model.

    6. occasionally run 'make_batch_files.py', eg. after svdaprep, so that model
    batch files are correct.  This should be run after moving the folder between
    platform as well.

To update/change a TOUGH2 model, the following files needs to be updated:

    - g_real_model.dat (model's mulgrid/geometry file)
    - real_model_original.dat (and .pdat)
    - real_model_original_pr.dat (and .pdat, rocktypes, geners not as important,
        they are processed from the natural state real_model_ogirinal.dat)
    - real_model.incon
    - and of course goPESTpar.list and goPESTobs.list
    - make sure to run 'python make_case_pst.py'

"""

if NUM_SLAVES is None:
    NUM_SLAVES = max(cpu_count() - 2, 1)
if PESTDIR:
    BEOPEST = os.path.join(PESTDIR,BEOPEST)

def ignore_dirs(folder, files):
    """ for shutil.copytree, ignores all subdirectories, ie. copy only files
    """
    ignore_list = []
    for f in files:
        full_path = os.path.join(folder, f)
        if os.path.isdir(full_path):
            ignore_list.append(f)
    return ignore_list

def run_cli(argv=[]):
    """ generate master and slave commands, copy files into slave directories,
    and launch them as process, wait until all finished.
    """
    if "pest_hp" in BEOPEST.lower():
        MASTER = BEOPEST
        if PESTDIR:
            AGENT = os.path.join(PESTDIR, cfg['pest']['executable_agent'])
        else:
            AGENT = cfg['pest']['executable_agent']
    else:
        MASTER = BEOPEST
        AGENT = BEOPEST

    generate_comm_files()
    print('Running BeoPEST with silent=', SILENT_SLAVES)

    # each command is a tuple (args, other options), see subproces.Popen()
    master = ([MASTER, PST_NAME, SWITCHES, '/h :%s' % PORT], {} )
    slaves, slaves_outs = [], []
    for i in range(NUM_SLAVES):
        s_dir = 'slave%i' % (i+1)
        if os.path.exists(s_dir):
            shutil.rmtree(s_dir)
        shutil.copytree('.', s_dir, ignore=ignore_dirs)
        if SILENT_SLAVES:
            # stdout_file = DEVNULL
            stdout_file = open(s_dir + '/stdout.txt', 'w')
            slaves_outs.append(stdout_file)
        else:
            stdout_file = None
        slaves.append((
            [AGENT, PST_NAME, '/h localhost:%s' % PORT],
            {
                'cwd': s_dir,
                'stdout': stdout_file,
                'stderr': stdout_file,
            }))

    # start master, sleep a bit, then start all slaves
    ps = []
    ps.append(Popen(master[0], **master[1]))
    time.sleep(0.1)
    for s in slaves:
        ps.append(Popen(s[0], **s[1]))
    for p in ps:
        p.wait()
    for f in slaves_outs:
        f.close()

def gen_run_management_file(nslave, slaves, wait=0.2, parlam=1, runtime=3600):
    """ generate run management file required by ppest/jactest.  I am
    implementing the simplified version here, single template and instruction
    file only.  TODO extend to be more flexible, which may require more info
    from pest control file.  Each slaves is a list of tuple (SLAVNAME,SLAVDIR),
    should have the same length as nslaves.
    """
    ifletyp = 0 # for now
    lines = []
    lines += [
        'prf',
        '%i %i %f %i' % (nslave, ifletyp, wait, parlam),
    ]
    for i in range(nslave):
        slavname, slavdir = slaves[i]
        if " " in slavname or "'" in slavname:
            slavname = slavname.replace("'", "''")
            slavname = "'" + slavname + "'"
        lines.append(slavname + ' ' + slavdir)
    lines.append(' '.join([str(runtime)] * nslave))
    # TODO: support non-zero IFLETYP
    return '\n'.join(lines)

def run_pslaves(master_command):
    """ generate master and slave commands, copy files into slave directories,
    and launch them as process, wait until all finished.  master_command should
    be a list of strings [cmd, arg1, arg2, ...].
    """
    # each command is a tuple (args, other options), see subproces.Popen()
    master = (master_command, {} )
    slaves = []
    for i in range(NUM_SLAVES):
        s_dir = 'slave%i' % (i+1)
        if os.path.exists(s_dir):
            shutil.rmtree(s_dir)
        shutil.copytree('.', s_dir, ignore=ignore_dirs)
        slaves.append((
            [BEOPEST, PST_NAME, '/h localhost:%s' % PORT],
            {
                'cwd': s_dir,
            }))

    # start master, sleep a bit, then start all slaves
    ps = []
    ps.append(Popen(master[0], **master[1]))
    time.sleep(0.1)
    for s in slaves:
        ps.append(Popen(s[0], **s[1]))
    for p in ps:
        p.wait()

def generate_comm_files():
    """ generate some files contain information so slaves know where things are.
    TODO: to improve, there should be better ways, partially to fit with my
    original NeSI system.
    """
    def write_to(filename, line):
        with open(filename, 'w') as f:
            f.write(line)
    write_to('_master_dir', os.getcwd())
    write_to('_pest_dir', PESTDIR)
    write_to('_tough2', TOUGH2)

if __name__ == '__main__':
    run_cli()



try:
    # works if calling from an repo or an pip install -e installation
    from setuptools_scm import get_version
    __version__ = get_version()
except LookupError:
    try:
        # works if calling from other pip install methods
        from ._version import version as __version__
    except ImportError:
        # worst case scenario
        __version__ = '0.0.0'



import unittest
import subprocess
import os

TESTDIR_EMPTY = './tests/data/test_empty_dir'

class TestCLI_NoConfig(unittest.TestCase):
    def setUp(self):
        self.original_dir = os.getcwd()
        os.chdir(TESTDIR_EMPTY)

    def tearDown(self):
        os.chdir(self.original_dir)

    def test_help_command(self):
        expected_output_components = [
            'Version: (',
            'Supported COMMANDs:',
            'help',
            'init [--no-copy][--no-par][--no-obs]',
            'submit',
            'run',
            'par',
            'obs',
            'run-pest-model',
            'run-forward',
            'save-iter-files',
            'check-slaves',
            'Important files for goPEST to work:',
            'goPESTconfig.toml',
            'goPESTpar.list',
            'goPESTobs.list',
            'University of Auckland, 2012, 2022',
        ]
        result = subprocess.run(['gopest'], capture_output=True, text=True)
        for exp in expected_output_components:
            self.assertIn(exp, result.stdout)
        result = subprocess.run(['gopest', 'help'], capture_output=True, text=True)
        for exp in expected_output_components:
            self.assertIn(exp, result.stdout)

if __name__ == '__main__':
    unittest.main()



import sys
import os
import time
import re
import json
import tarfile
import subprocess
from shutil import copy2, move
from time import sleep
import importlib.util
import inspect

import numpy as np
import h5py

from mulgrids import *
from t2data import *
from t2listing import *
from t2incons import *

from gopest.utils import gener_groups

from gopest.common import config
from gopest.common import runtime

def tail(filename, n=10):
    """ Return the last n lines of a file """
    from collections import deque
    return deque(open(filename), n)

def read_gener_file(fname):
    f = t2data_parser(fname,'rU')
    line = f.readline()
    if line[:5] != 'GENER':
        # get rid of first line 'GENER' and check
        raise Exception("File %s is not a valid generator file, must starts with GENER." % fname)
    gs = t2data()
    gs.read_generators(f)
    f.close()
    return gs.generatorlist

def read_cols_in_zones(geo, fname='real_model.json'):
    """
    NOTE WIP, need to deal with zone order in some cases (one zone includes another)
    """
    with open(fname, 'rU') as jf:
        cfg = json.load(jf)
        col_zone = {}
        for zone in cfg['ZonePolygon'].keys():
            zone_polygon = [np.array(pt) for pt in cfg['ZonePolygon'][zone]]
            for c in geo.columns_in_polygon(zone_polygon):
                col_zone[c.name] = zone
    return col_zone

def save_bad_model(dat, sav):
    """ keep a model and its incon aside for later inspection, iused in the
    case of difficult to converge.
    """
    archive_dir = 'bad_models'
    def next_unused_dir():
        from os.path import isdir, isfile, join
        i = 1
        p = join(archive_dir, str(i))
        while isdir(p) or isfile(p):
            i += 1
            p = join(archive_dir, str(i))
        return p
    psave = next_unused_dir()
    os.makedirs(psave)
    dat.write(psave + os.sep + 'real_model.dat')
    sav.write(psave + os.sep + 'real_model.incon')

def targz_files(files, default_name):
    """ add files into a compressed tar file (.tar.gz).  File name will be
    default_name with a number.  The number will be incremented if other tar
    file existed based on the same name.
    """
    i = 1
    f = default_name + str(i) + ".tar.gz"
    while os.path.isfile(f):
        i += 1
        f = default_name + str(i) + ".tar.gz"
    tar = tarfile.open(f, 'w:gz')
    for name in files:
        if os.path.isfile(name) or os.path.isdir(name):
            tar.add(name)
    tar.close()
    return f

def fix_base_heat(dat, geo):
    fname = 'real_model.json'
    col_zone = read_cols_in_zones(geo, fname=fname)
    with open(fname, 'r') as jf:
        cfg = json.load(jf)
    for g in dat.generatorlist:
        if g.name.endswith('99') and g.type == 'HEAT':
            col = geo.column[geo.column_name(g.block)]
            if col.name in col_zone:
                v = cfg['HeatFlux'][col_zone[col.name]]
            else:
                v = cfg['HeatFlux']['Default']
            g.gx = col.area * v
    return dat

def fix_co2_mass_ratio(dat):
    # ohaaki seems to use 4%, have not checked with Mike yet
    # CO2_MASS_RATIO = 0.04
    with open('data_co2_ratio.json', 'r') as f:
        ratio = json.load(f)
    mass = {}
    for g in dat.generatorlist:
        if g.block.endswith('61') and g.name.endswith('77') and g.type == 'MASS':
            mass[g.block] = g.gx
    for g in dat.generatorlist:
        if g.block.endswith('61') and g.name.endswith('66') and g.type == 'COM2':
            # g.gx = mass[g.block] * CO2_MASS_RATIO
            g.gx = mass[g.block] * ratio[g.block]
    return dat

def set_output_times(dat, print_every=31558000.0):
    """ a year is actually 31557600 sec, but precision will be lost, so it would
    be 31558000 that actually used. """
    dat.output_times.update({
        'max_timestep': 0.0,
        'num_times': 1000,
        'num_times_specified': 1,
        'time': [print_every],
        'time_increment': print_every,
        })
    dat.parameter.update({
        'max_timesteps': -999, # unlimited
        'print_interval': -999, # do not print with MCYPR
        })
    dat.parameter['option'][24] = 2
    dat.parameter['option'][16] = 5
    return dat

def ensure_converge(ns, sav, allow_failed_ns=True):
    # re-run until taget time reached, max 3 times.
    from os.path import splitext
    datbase,ext = splitext(ns.filename)
    target_ns_time = ns.parameter['tstop']
    run = 0
    while abs(sav.timing['sumtim'] - target_ns_time) > 1000 and run < 0:
        run += 1
        print('Warning NS ends at %e sec, did not reach target time: %e, re-run NS for %ith time.' % (float(sav.timing['sumtim']), float(target_ns_time), run))
        ns.write(datbase+"_rerun.dat", echo_extra_precision=True)
        sav.write(datbase+"_rerun.incon", reset=True)
        ns.run(simulator=simulator, silent=silent)
        sav = t2incon(datbase+"_rerun.save")
    if abs(sav.timing['sumtim'] - target_ns_time) > 1000:
        print('Warning NS ends at %e sec, still not reach target time: %e, after %ith tries.' % (float(sav.timing['sumtim']), float(target_ns_time), run))
        tfname = targz_files([datbase+'.dat', datbase+'.pdat', datbase+'.incon'],
            'bad_model_')
        print(tfname, 'saved')
        if not allow_failed_ns:
            return False
    return sav

def del_files_no_check(files):
    for cf in files:
        try:
            os.remove(cf)
        except OSError:
            pass

def run_user_pre(seq):
    """ run user supplied pre-processing function if exists """
    file_path = 'goPESTuser.py'
    module_name = 'user'
    if os.path.exists(file_path):
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        for name,func in inspect.getmembers(module, inspect.isfunction):
            print('%s has user func: %s()' % (file_path, name))
            if name == 'pre_' + seq:
                print('  -> calling %s() ...' % name)
                func()
                break

def run_ns_pr_waiwera(skippr=False, sav2inc=False, simulator='waiwera-dkr',
              allow_failed_ns=True, silent=True):
    """
    supported platforms:
        'waiwera' - local native waiwera executable
        'waiwera-dkr' - local waiwera running on docker
        'waiwera-Maui' - NeSI Maui (uses submit_beopest.py)
        'waiwera-Mahuika' - NeSI Mahuika (uses submit_beopest.py)
    """
    fgeo = runtime['filename']['geom']
    fsav = runtime['filename']['save']
    finc = runtime['filename']['incon']
    fdato = runtime['filename']['dat_orig']
    fdats = runtime['filename']['dat_seq']
    flsts = runtime['filename']['lst_seq']
    sequence = config['model']['sequence']

    if os.path.exists(finc):
        # check incon
        inc_h5 = h5py.File(finc, 'r')
        inc_idx = len(inc_h5['time'][:,0]) - 1
        if inc_idx < 0:
            inc_h5.close()
            raise Exception("ERROR! Waiwera initial conditions file '%s' has no data." % finc)
        inc_h5.close()
    else:
        raise Exception("Error! Waiwera initial conditions file '%s' not found." % finc)

    # call user pre-processing
    # user is responsible of reading/writing files
    run_user_pre(sequence[0])

    with open(fdats[0], 'r') as f:
        wai_ns = json.load(f)
    # overwrite these just to be safe
    wai_ns["output"]["filename"] = flsts[0]
    wai_ns["mesh"]["filename"] = "g_real_model.msh"
    wai_ns["initial"] = {"filename": finc, "index": inc_idx}
    if silent:
        wai_ns["logfile"] = {"echo": False}
    else:
        wai_ns["logfile"] = {"echo": True}
    with open(fdats[0], 'w') as f:
        json.dump(wai_ns, f, indent=2, sort_keys=True)

    # clean up
    del_files_no_check([
        flsts[0],
        fdats[0].replace('.json', '.yaml'),
        ])

    model_args = [fdats[0]] + config['simulator']['cmd-options']
    if config['mode'] == 'local':
        cmd = [config['simulator']['executable']] + model_args
    else:
        cmd = ['gopest', 'submit', '--forward3x', ' '.join(model_args)]

    START_TIME = time.time()
    print('NS launched on %s...' % config['mode'])
    # run NS
    print(cmd)
    subprocess.call(cmd)

    # check NS
    if os.path.exists(flsts[0]):
        inc_h5 = h5py.File(flsts[0], 'r')
        inc_idx = len(inc_h5['time'][:,0]) - 1
        if inc_idx < 0:
            raise Exception("ERROR! output file '%s' has no data." % flsts[0])
        endtime = inc_h5['time'][inc_idx, 0]
        inc_h5.close()
        if abs(endtime - wai_ns['time']['stop']) < 1.e3:
            print('NS finished after %.1f seconds' % (time.time() - START_TIME))
        else:
            print('NS failed after %.1f seconds' % (time.time() - START_TIME))
            if not allow_failed_ns:
                return False
    else:
        raise Exception('NS failed, result %s not found.' % flsts[0])

    if skippr:
        msg = 'Skipped PR, use NS result as PR for goPESTobs/pest_model'
        print(msg)
        return True

    # call user pre-processing
    # user is responsible of reading/writing files
    run_user_pre(sequence[-1])

    with open(fdats[-1], 'r') as f:
        wai_pr = json.load(f)
    wai_pr['output']['filename'] = flsts[-1]
    wai_pr["initial"] = {"filename": flsts[0], "index": inc_idx}
    wai_pr["mesh"]["filename"] = "g_real_model.msh"
    if silent:
        wai_pr["logfile"] = {"echo": False}
    else:
        wai_pr["logfile"] = {"echo": True}
    with open(fdats[-1], 'w') as f:
        json.dump(wai_pr, f, indent=2, sort_keys=True)

    # clean up
    del_files_no_check([
        flsts[-1],
        fdats[-1].replace('.json', '.yaml'),
        ])

    model_args = [fdats[-1]] + config['simulator']['cmd-options']
    if config['mode'] == 'local':
        cmd = [config['executable']] + model_args
    else:
        cmd = ['gopest', 'submit', '--forward3x', ' '.join(model_args)]

    START_TIME = time.time()
    print('PR launched on %s...' % simulator)
    # run
    print(cmd)
    subprocess.call(cmd)

    # check PR
    if os.path.exists(flsts[-1]):
        inc_h5 = h5py.File(flsts[-1], 'r')
        inc_idx = len(inc_h5['time'][:,0]) - 1
        if inc_idx < 0:
            raise Exception("ERROR! output file '%s' has no data." % flsts[-1])
        endtime = inc_h5['time'][inc_idx, 0]
        inc_h5.close()
        if abs(endtime - wai_pr['time']['stop']) < 1.e3:
            print('PR finished after %.1f seconds' % (time.time() - START_TIME))
        else:
            print('PR failed after %.1f seconds' % (time.time() - START_TIME))
            if not allow_failed_ns:
                return False
    else:
        raise Exception('PR failed, result %s not found.' % flsts[-1])

    return True

def run_ns_pr_mixed(skippr=False, sav2inc=False, simulator='waiwera-dkr',
              allow_failed_ns=True, silent=True):
    """ convert aut2 model into waiwera, and run ns/pr

    supported platforms:
        'waiwera' - local native waiwera executable
        'waiwera-dkr' - local waiwera running on docker
        'waiwera-Maui' - NeSI Maui (uses submit_beopest.py)
        'waiwera-Mahuika' - NeSI Mahuika (uses submit_beopest.py)
    """
    geo = mulgrid("g_real_model.dat")
    inc = t2incon("real_model.incon")
    inc.porosity = None
    inc.write("real_model.incon", reset=True)
    inc = t2incon("real_model.incon")
    t2_ns = t2data("real_model.dat")
    t2_ns = fix_co2_mass_ratio(t2_ns)
    t2_ns.write(echo_extra_precision=True)

    if os.path.exists('real_model.incon.h5'):
        # check incon
        inc_h5 = h5py.File('real_model.incon.h5', 'r')
        inc_idx = len(inc_h5['time'][:,0]) - 1
        if inc_idx < 0:
            raise Exception("ERROR! Waiwera initial conditions file 'real_model.incon.h5' has no data.")
        inc_h5.close()
        use_inc = '--' # use dummy in dat.json(), then replaced later
    else:
        print('real_model.incon.h5 does not exist, use T2 real_model.incon instead.')
        use_inc = inc

    wai_ns = t2_ns.json(geo, 'g_real_model.msh',
                     atmos_volume=1.e25,
                     incons=use_inc,
                     eos=None,
                     bdy_incons=inc, # somehow I still need it to avoid error
                     mesh_coords='xyz')

    wai_ns_orig = json.load(open('real_model_original.json', 'r'))
    for k in ["time", "output", "mesh"]:
        wai_ns[k] = wai_ns_orig[k]

    # overwrite these just to be safe
    wai_ns["thermodynamics"] = {"name": "ifc67", "extrapolate": True}
    wai_ns["output"]["filename"] = "real_model.h5"
    wai_ns["output"]["frequency"] = 1000
    wai_ns["time"]["step"]["maximum"]["number"] = 80000
    wai_ns["time"]["stop"] = 0.5e14
    wai_ns["mesh"]["filename"] = "g_real_model.msh"
    if use_inc == '--':
       wai_ns["initial"] = {"filename": "real_model.incon.h5", "index": inc_idx}

    json.dump(wai_ns, open('real_model.json', 'w'), indent=2, sort_keys=True)

    # clean up
    clean_files = [
        'real_model.h5',
        'real_model.save',
        'real_model.listing',
    ]
    for cf in clean_files:
        try:
            os.remove(cf)
        except OSError:
            pass
    model_cmd = [
        'real_model.json',
        '-ksp_type', 'bcgsl',
        '-snes_ksp_ew',
        '-snes_ksp_ew_rtol0', '1e-5',
        '-snes_ksp_ew_rtolmax', '1e-4',
        '-pc_type', 'asm',
        '-pc_factor_levels', '1',
        '-snes_max_linear_solve_fail', '3',
        ]
    NP = 4
    cmd = {
        'waiwera': ['mpiexec', '-np', str(NP), 'waiwera', '-np'] + model_cmd,
        'waiwera-dkr': ['waiwera-dkr', '-np', str(NP), '--tag', 'testing'] + model_cmd,
        'waiwera-Maui': ['python', 'submit_beopest.py', '--forward3maui', ' '.join(['~/bin/waiwera-maui']+model_cmd)],
        'waiwera-Mahuika': ['python', 'submit_beopest.py', '--forward3mahuika', ' '.join(['~/bin-mahuika/waiwera']+model_cmd)],
    }[simulator]
    START_TIME = time.time()
    print('NS launched on %s...' % simulator)
    # run NS
    print(cmd)
    subprocess.call(cmd)

    # check NS
    if os.path.exists('real_model.h5'):
        inc_h5 = h5py.File('real_model.h5', 'r')
        inc_idx = len(inc_h5['time'][:,0]) - 1
        if inc_idx < 0:
            raise Exception("ERROR! output file 'real_model.h5' has no data.")
        endtime = inc_h5['time'][inc_idx, 0]
        inc_h5.close()
        if abs(endtime - wai_ns['time']['stop']) < 1.e3:
            print('NS finished after %.1f seconds' % (time.time() - START_TIME))
        else:
            print('NS failed after %.1f seconds' % (time.time() - START_TIME))
            if not allow_failed_ns:
                return False

    if skippr:
        copy2('real_model.h5', 'real_model_pr.h5')
        msg = 'Skipped PR, use NS result as PR for goPESTobs/pest_model'
        print(msg)
        return True

    # change NS to PR
    wai_pr = wai_ns
    with open('gs_production.json', 'r') as f:
        data = json.load(f)
        wai_pr['source'] = wai_pr['source'] + data
    with open('real_model_pr.output.json', 'r') as f:
        data = json.load(f)
        wai_pr['output'] = data
    with open('real_model_pr.time.json', 'r') as f:
        data = json.load(f)
        wai_pr['time'] = data
    # additional
    wai_pr['output']['filename'] = 'real_model_pr.h5'
    wai_pr['output']['frequency'] = 0
    wai_pr["initial"] = {"filename": "real_model.h5", "index": inc_idx}

    json.dump(wai_pr, open('real_model_pr.json', 'w'), indent=2, sort_keys=True)

    # clean up
    clean_files = [
        'real_model_pr.h5',
        'real_model_pr.save',
        'real_model_pr.listing',
    ]
    for cf in clean_files:
        try:
            os.remove(cf)
        except OSError:
            pass
    START_TIME = time.time()
    print('PR launched on %s...' % simulator)
    model_cmd = [
        'real_model_pr.json',
        '-ksp_type', 'bcgsl',
        '-snes_ksp_ew',
        '-snes_ksp_ew_rtol0', '1e-5',
        '-snes_ksp_ew_rtolmax', '1e-4',
        '-pc_type', 'asm',
        '-pc_factor_levels', '1',
        '-snes_max_linear_solve_fail', '3',
        ]
    NP = 4
    cmd = {
        'waiwera': ['mpiexec', '-np', str(NP), 'waiwera', '-np'] + model_cmd,
        'waiwera-dkr': ['waiwera-dkr', '-np', str(NP), '--tag', 'testing'] + model_cmd,
        'waiwera-Maui': ['python', 'submit_beopest.py', '--forward3maui', ' '.join(['~/bin/waiwera-maui']+model_cmd)],
        'waiwera-Mahuika': ['python', 'submit_beopest.py', '--forward3mahuika', ' '.join(['~/bin-mahuika/waiwera']+model_cmd)],
    }[simulator]
    # run
    print('Running PR ...')
    print(cmd)
    START_TIME = time.time()
    subprocess.call(cmd)

    # check PR
    if os.path.exists('real_model_pr.h5'):
        inc_h5 = h5py.File('real_model_pr.h5', 'r')
        inc_idx = len(inc_h5['time'][:,0]) - 1
        if inc_idx < 0:
            raise Exception("ERROR! output file 'real_model.h5' has no data.")
        endtime = inc_h5['time'][inc_idx, 0]
        inc_h5.close()
        if abs(endtime - wai_pr['time']['stop']) < 1.e3:
            print('PR finished after %.1f seconds' % (time.time() - START_TIME))
        else:
            print('PR failed after %.1f seconds' % (time.time() - START_TIME))
            if not allow_failed_ns:
                return False
    # fake a dummy real_model_pr.dat (for goPESTobs.py), need to check not used by those obs types
    t2_ns.write('real_model_pr.dat', echo_extra_precision=True)
    return True

def run_ns_pr_aut2(skippr=False, sav2inc=False, simulator='AUTOUGH2_3D',
                   allow_failed_ns=True, silent=True):
    # fgeo = runtime['filename']['geom']
    finc = runtime['filename']['incon']
    fdats = runtime['filename']['dat_seq']
    flsts = runtime['filename']['lst_seq']
    fincs = runtime['filename']['inc_seq']
    fsavs = runtime['filename']['sav_seq']
    sequence = config['model']['sequence']

    silent = config['model']['silent']
    print('   -- aut2: %s is used.  (silent=%s)' % (simulator, str(silent)))

    # geo = mulgrid("g_real_model.dat")
    sav = t2incon(finc)

    for iseq,seq in enumerate(sequence):
        sav.porosity = None
        sav.write(fincs[iseq], reset=True)
        inc = t2incon(fincs[iseq])

        # call user pre-processing
        # user is responsible of reading/writing files
        run_user_pre(seq)

        ns = t2data(fdats[iseq])
        ns.write(echo_extra_precision=True)

        START_TIME = time.time()
        ns.run(simulator=simulator, silent=silent)
        sav = t2incon(fsavs[iseq])
        if seq == 'ns':
            sav = ensure_converge(ns, sav, allow_failed_ns)
            if sav is False:
                return False
            if sav2inc:
                sav.write("real_model.incon", reset=True)
        print('    - %s finished after ' % seq, (time.time() - START_TIME), 'seconds')

    # if skippr:
    #     # fake real_model_pr.* because thats what pest_model.py will apply goPESTobs on
    #     copy2('real_model.dat', 'real_model_pr.dat')
    #     copy2('real_model.pdat', 'real_model_pr.pdat')
    #     copy2('real_model.listing', 'real_model_pr.listing')
    #     msg = 'Skipped PR, make sure to use NS result as PR for goPESTobs/pest_model'
    #     print(msg)
    #     return True

    # ##### everything below is production run
    # pr = t2data("real_model_original_pr.dat")
    # target_pr_time = pr.parameter['tstop']
    # # pr = set_output_times(pr, 15779000.0) # every half year
    # pr = set_output_times(pr, 31558000.0) # every year

    # pr.grid = ns.grid
    # pr.clear_generators()
    # for g in ns.generatorlist:
    #     pr.add_generator(g)
    # # Append all additinal generators at the end
    # gs_to_add = read_gener_file('production.geners')
    # for g in gs_to_add:
    #     pr.add_generator(g)

    # pr.write("real_model_pr.dat", echo_extra_precision=True)
    # sav.write("real_model_pr.incon", reset=True)

    # print('Running PR ...')
    # START_TIME = time.time()
    # pr.run(simulator=simulator, silent=silent)
    # print('PR finished after ', (time.time() - START_TIME), 'seconds')

    # prsav = t2incon("real_model_pr.save")
    # if abs(prsav.timing['sumtim'] - target_pr_time) > 1000:
    #     print('Warning PR ends at %e sec, did not reach target time: %e' % (float(prsav.timing['sumtim']), float(target_pr_time)))
    #     print(''.join([line for line in tail('real_model_pr.listing', n=20)]))

    return True

def get_t2():
    with open('_tough2', 'r') as f:
        line = f.readlines()[0].strip()
        return line

def run_ns_pr():
    """ call this from pest_model.py """
    if config['simulator']['input-type'] == 'waiwera':
        return run_ns_pr_waiwera(skippr=config['model']['skip-pr'],
                                 sav2inc=True,
                                 simulator=config['simulator']['executable'],
                                 allow_failed_ns=True,
                                 silent=False)
    elif config['simulator']['input-type'] == 'aut2':
        return run_ns_pr_aut2(skippr=config['model']['skip-pr'],
                              sav2inc=False,
                              simulator=config['simulator']['executable'],
                              allow_failed_ns=True,
                              silent=False)
    else:
        raise NotImplementedError('No implementation of run_ns_pr() for input-type: ' + config['simulator']['input-type'])

def main_cli(argv=[]):
    skippr = config['model']['skip-pr']
    sav2inc = False

    if len(argv) > 1:
        if '--skip-pr' in argv[1:]:
            skippr = True
        if '--sav2inc' in argv[1:]:
            sav2inc = True

    if config['simulator']['input-type'] == 'waiwera':
        run_ns_pr_waiwera(skippr, sav2inc, simulator=get_t2(), silent=True)
    else:
        run_ns_pr_aut2(skippr=False, sav2inc=False,
                       simulator=config['simulator']['executable'],
                       allow_failed_ns=True, silent=True)
